# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import os
import re
import sys
from io import StringIO
from typing import TYPE_CHECKING
from typing import Any
from unittest import mock

from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import log
from twisted.python.filepath import FilePath
from twisted.trial import unittest
from zope.interface import implementer

from buildbot.config.master import MasterConfig
from buildbot.data import resultspec
from buildbot.interfaces import IConfigLoader
from buildbot.master import BuildMaster
from buildbot.plugins import worker
from buildbot.process.properties import Interpolate
from buildbot.process.results import SUCCESS
from buildbot.process.results import statusToString
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util.misc import DebugIntegrationLogsMixin
from buildbot.test.util.sandboxed_worker import SandboxedWorker
from buildbot.util.twisted import any_to_async
from buildbot.util.twisted import async_to_deferred
from buildbot.worker.local import LocalWorker
from buildbot_worker.bot import Worker

if TYPE_CHECKING:
    from twisted.application.service import Service


@implementer(IConfigLoader)
class DictLoader:
    def __init__(self, config_dict):
        self.config_dict = config_dict

    def loadConfig(self):
        return MasterConfig.loadFromDict(self.config_dict, '<dict>')


class TestedMaster:
    def __init__(self) -> None:
        self.master: BuildMaster | None = None
        self.is_master_shutdown = False

    @async_to_deferred
    async def create_master(self, case, reactor, config_dict, basedir=None):
        """
        Create a started ``BuildMaster`` with the given configuration.
        """
        if basedir is None:
            basedir = case.mktemp()
        os.makedirs(basedir, exist_ok=True)
        config_dict['buildbotNetUsageData'] = None
        self.master = BuildMaster(basedir, reactor=reactor, config_loader=DictLoader(config_dict))

        if 'db_url' not in config_dict:
            config_dict['db_url'] = 'sqlite://'

        # TODO: Allow BuildMaster to transparently upgrade the database, at least
        # for tests.
        self.master.config.db.db_url = config_dict['db_url']
        await self.master.db.setup(check_version=False)
        await self.master.db.model.upgrade()
        self.master.db.setup = lambda: None

        await self.master.startService()

        case.addCleanup(self.shutdown)

        return self.master

    @async_to_deferred
    async def shutdown(self):
        if self.is_master_shutdown:
            return
        try:
            await self.master.stopService()
        except Exception as e:
            log.err(e)
        try:
            await self.master.db.pool.stop()
        except Exception as e:
            log.err(e)
        self.is_master_shutdown = True


def print_test_log(l, out):
    print(" " * 8 + f"*********** LOG: {l['name']} *********", file=out)
    if l['type'] == 's':
        for line in l['contents']['content'].splitlines():
            linetype = line[0]
            line = line[1:]
            if linetype == 'h':
                # cyan
                line = "\x1b[36m" + line + "\x1b[0m"
            if linetype == 'e':
                # red
                line = "\x1b[31m" + line + "\x1b[0m"
            print(" " * 8 + line)
    else:
        print("" + l['contents']['content'], file=out)
    print(" " * 8 + "********************************", file=out)


@async_to_deferred
async def enrich_build(
    build, master: BuildMaster, want_steps=False, want_properties=False, want_logs=False
):
    # enrich the build result, with the step results
    if want_steps:
        build["steps"] = await master.data.get(("builds", build['buildid'], "steps"))
        # enrich the step result, with the logs results
        if want_logs:
            build["steps"] = list(build["steps"])
            for step in build["steps"]:
                step['logs'] = await master.data.get(("steps", step['stepid'], "logs"))
                step["logs"] = list(step['logs'])
                for l in step["logs"]:
                    l['contents'] = await master.data.get((
                        "logs",
                        l['logid'],
                        "contents",
                    ))

    if want_properties:
        build["properties"] = await master.data.get((
            "builds",
            build['buildid'],
            "properties",
        ))


@async_to_deferred
async def print_build(build, master: BuildMaster, out=sys.stdout, with_logs=False):
    # helper for debugging: print a build
    await enrich_build(build, master, want_steps=True, want_properties=True, want_logs=True)
    print(
        f"*** BUILD {build['buildid']} *** ==> {build['state_string']} "
        f"({statusToString(build['results'])})",
        file=out,
    )
    for step in build['steps']:
        print(
            f"    *** STEP {step['name']} *** ==> {step['state_string']} "
            f"({statusToString(step['results'])})",
            file=out,
        )
        for url in step['urls']:
            print(f"       url:{url['name']} ({url['url']})", file=out)
        for l in step['logs']:
            print(f"        log:{l['name']} ({l['num_lines']})", file=out)
            if step['results'] != SUCCESS or with_logs:
                print_test_log(l, out)


class RunFakeMasterTestCase(TestReactorMixin, DebugIntegrationLogsMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        self.setupDebugIntegrationLogs()

        def cleanup():
            self.assertFalse(self.master.running, "master is still running!")

        self.addCleanup(cleanup)

    @defer.inlineCallbacks
    def setup_master(self, config_dict):
        self.tested_master = TestedMaster()
        self.master = yield self.tested_master.create_master(self, self.reactor, config_dict)

    @defer.inlineCallbacks
    def reconfig_master(self, config_dict=None):
        if config_dict is not None:
            self.master.config_loader.config_dict = config_dict
        yield self.master.doReconfig()

    @defer.inlineCallbacks
    def clean_master_shutdown(self, quick=False):
        yield self.master.botmaster.cleanShutdown(quickMode=quick, stopReactor=False)

    def createLocalWorker(self, name, **kwargs):
        workdir = FilePath(self.mktemp())
        workdir.createDirectory()
        return LocalWorker(name, workdir.path, **kwargs)

    @defer.inlineCallbacks
    def assertBuildResults(self, build_id, result):
        dbdict = yield self.master.db.builds.getBuild(build_id)
        self.assertEqual(result, dbdict.results)

    @defer.inlineCallbacks
    def assertStepStateString(self, step_id, state_string):
        datadict = yield self.master.data.get(('steps', step_id))
        self.assertEqual(datadict['state_string'], state_string)

    @defer.inlineCallbacks
    def assertLogs(self, build_id, exp_logs):
        got_logs = {}
        data_logs = yield self.master.data.get(('builds', build_id, 'steps', 1, 'logs'))
        for l in data_logs:
            self.assertTrue(l['complete'])
            log_contents = yield self.master.data.get((
                'builds',
                build_id,
                'steps',
                1,
                'logs',
                l['slug'],
                'contents',
            ))

            got_logs[l['name']] = log_contents['content']

        self.assertEqual(got_logs, exp_logs)

    @defer.inlineCallbacks
    def create_build_request(self, builder_ids, properties=None):
        properties = properties.asDict() if properties is not None else None
        ret = yield self.master.data.updates.addBuildset(
            waited_for=False,
            builderids=builder_ids,
            sourcestamps=[
                {'codebase': '', 'repository': '', 'branch': None, 'revision': None, 'project': ''},
            ],
            properties=properties,
        )
        # run debounced calls
        self.master.reactor.advance(1)
        return ret

    @defer.inlineCallbacks
    def do_test_build_by_name(self, builder_name):
        builder_id = yield self.master.data.updates.findBuilderId(builder_name)
        yield self.do_test_build(builder_id)

    @defer.inlineCallbacks
    def do_test_build(self, builder_id):
        # setup waiting for build to finish
        d_finished = defer.Deferred()

        def on_finished(_, __):
            if not d_finished.called:
                d_finished.callback(None)

        consumer = yield self.master.mq.startConsuming(on_finished, ('builds', None, 'finished'))

        # start the builder
        yield self.create_build_request([builder_id])

        # and wait for build completion
        yield d_finished
        yield consumer.stopConsuming()


class TestedRealMaster(TestedMaster):
    def __init__(self) -> None:
        super().__init__()
        self.case: unittest.TestCase | None = None
        self.worker: Service | None = None

    @async_to_deferred
    async def setup_master(
        self,
        case: unittest.TestCase,
        reactor,
        config_dict,
        proto='null',
        basedir=None,
        start_worker=True,
        **worker_kwargs,
    ):
        """
        Setup and start a master configured by config_dict
        """

        self.case = case

        # mock reactor.stop (which trial *really* doesn't like test code to call!)
        stop = mock.create_autospec(reactor.stop)
        case.patch(reactor, 'stop', stop)

        if start_worker:
            config_protocols: dict[str, Any] = {}
            if proto == 'pb':
                config_protocols = {"pb": {"port": "tcp:0:interface=127.0.0.1"}}
                workerclass = worker.Worker
            elif proto == 'msgpack':
                config_protocols = {"msgpack_experimental_v7": {"port": 0}}
                workerclass = worker.Worker
            elif proto == 'null':
                config_protocols = {"null": {}}
                workerclass = worker.LocalWorker
            else:
                raise RuntimeError(f"{proto} protocol is not supported.")
            config_dict['workers'] = [
                workerclass("local1", password=Interpolate("localpw"), missing_timeout=0)
            ]
            config_dict['protocols'] = config_protocols

        await self.create_master(case, reactor, config_dict, basedir=basedir)
        self.master_config_dict = config_dict
        case.assertFalse(stop.called, "startService tried to stop the reactor; check logs")

        if not start_worker:
            return

        assert self.master
        if proto in ('pb', 'msgpack'):
            sandboxed_worker_path = os.environ.get("SANDBOXED_WORKER_PATH", None)
            if proto == 'pb':
                protocol = 'pb'
                dispatcher = next(iter(self.master.pbmanager.dispatchers.values()))
            else:
                protocol = 'msgpack_experimental_v7'
                dispatcher = next(iter(self.master.msgmanager.dispatchers.values()))

                # We currently don't handle connection closing cleanly.
                dispatcher.serverFactory.setProtocolOptions(closeHandshakeTimeout=0)

            worker_port = dispatcher.port.getHost().port

            # create a worker, and attach it to the master, it will be started, and stopped
            # along with the master
            worker_dir = FilePath(case.mktemp())
            worker_dir.createDirectory()
            if sandboxed_worker_path is None:
                self.worker = Worker(
                    "127.0.0.1",
                    worker_port,
                    "local1",
                    "localpw",
                    worker_dir.path,
                    False,
                    protocol=protocol,
                    **worker_kwargs,
                )
            else:
                self.worker = SandboxedWorker(
                    "127.0.0.1",
                    worker_port,
                    "local1",
                    "localpw",
                    worker_dir.path,
                    sandboxed_worker_path,
                    protocol=protocol,
                    **worker_kwargs,
                )

        if self.worker is not None:
            await any_to_async(self.worker.setServiceParent(self.master))

        case.addCleanup(self.dump_data_if_failed)

    @async_to_deferred
    async def shutdown(self):
        if self.is_master_shutdown:
            return

        if isinstance(self.worker, SandboxedWorker):
            try:
                await self.worker.shutdownWorker()
            except Exception as e:
                log.err(e)
        await super().shutdown()

    @async_to_deferred
    async def dump_data_if_failed(self):
        if self.case is not None and not self.case._passed and not self.is_master_shutdown:
            dump = StringIO()
            print("FAILED! dumping build db for debug", file=dump)
            builds = await self.master.data.get(("builds",))
            for build in builds:
                await print_build(build, self.master, dump, with_logs=True)

            raise self.case.failureException(dump.getvalue())


class RunMasterBase(unittest.TestCase):
    proto = "null"

    # All tests that start master need higher timeout due to test runtime variability on
    # oversubscribed hosts.
    timeout = 60

    @defer.inlineCallbacks
    def setup_master(self, config_dict, startWorker=True, basedir=None, **worker_kwargs):
        self.tested_master = TestedRealMaster()
        yield self.tested_master.setup_master(
            self,
            reactor,
            config_dict,
            proto=self.proto,
            basedir=basedir,
            start_worker=startWorker,
            **worker_kwargs,
        )
        self.master = self.tested_master.master
        self.master_config_dict = self.tested_master.master_config_dict

    @defer.inlineCallbacks
    def doForceBuild(
        self,
        wantSteps=False,
        wantProperties=False,
        wantLogs=False,
        useChange=False,
        forceParams=None,
        triggerCallback=None,
    ):
        if forceParams is None:
            forceParams = {}
        # force a build, and wait until it is finished
        d = defer.Deferred()

        # in order to allow trigger based integration tests
        # we wait until the first started build is finished
        self.firstbsid = None

        def newCallback(_, data):
            if self.firstbsid is None:
                self.firstbsid = data['bsid']
                newConsumer.stopConsuming()

        def finishedCallback(_, data):
            if self.firstbsid == data['bsid']:
                d.callback(data)

        newConsumer = yield self.master.mq.startConsuming(newCallback, ('buildsets', None, 'new'))

        finishedConsumer = yield self.master.mq.startConsuming(
            finishedCallback, ('buildsets', None, 'complete')
        )

        if triggerCallback is not None:
            yield triggerCallback()
        elif useChange is False:
            # use data api to force a build
            yield self.master.data.control("force", forceParams, ("forceschedulers", "force"))
        else:
            # use data api to force a build, via a new change
            yield self.master.data.updates.addChange(**useChange)

        # wait until we receive the build finished event
        buildset = yield d
        buildrequests = yield self.master.data.get(
            ('buildrequests',), filters=[resultspec.Filter('buildsetid', 'eq', [buildset['bsid']])]
        )
        buildrequest = buildrequests[-1]
        builds = yield self.master.data.get(
            ('builds',),
            filters=[resultspec.Filter('buildrequestid', 'eq', [buildrequest['buildrequestid']])],
        )
        # if the build has been retried, there will be several matching builds.
        # We return the last build
        build = builds[-1]
        finishedConsumer.stopConsuming()
        yield enrich_build(build, self.master, wantSteps, wantProperties, wantLogs)
        return build

    def _match_patterns_consume(self, text, patterns, is_regex):
        for pattern in patterns[:]:
            if is_regex:
                if re.search(pattern, text):
                    patterns.remove(pattern)
            else:
                if pattern in text:
                    patterns.remove(pattern)
        return patterns

    @defer.inlineCallbacks
    def checkBuildStepLogExist(self, build, expectedLog, onlyStdout=False, regex=False):
        if isinstance(expectedLog, str):
            expectedLog = [expectedLog]
        if not isinstance(expectedLog, list):
            raise RuntimeError(
                'The expectedLog argument must be either string or a list of strings'
            )

        yield enrich_build(
            build, self.master, want_steps=True, want_properties=True, want_logs=True
        )
        for step in build['steps']:
            for l in step['logs']:
                for line in l['contents']['content'].splitlines():
                    if onlyStdout and line[0] != 'o':
                        continue
                    expectedLog = self._match_patterns_consume(line, expectedLog, is_regex=regex)
        if expectedLog:
            print(f"{expectedLog} not found in logs")
        return len(expectedLog) == 0



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial.unittest import SynchronousTestCase

from buildbot.test.util.patch_delay import patchForDelay


class TestException(Exception):
    pass


def fun_to_patch(*args, **kwargs):
    return defer.succeed((args, kwargs))


def fun_to_patch_exception():
    raise TestException()


non_callable = 1


class Tests(SynchronousTestCase):
    def test_raises_not_found(self):
        with self.assertRaises(RuntimeError):
            with patchForDelay(__name__ + '.notfound'):
                pass

    def test_raises_not_callable(self):
        with self.assertRaises(RuntimeError):
            with patchForDelay(__name__ + '.non_callable'):
                pass

    def test_patches_within_context(self):
        d = fun_to_patch()
        self.assertTrue(d.called)

        with patchForDelay(__name__ + '.fun_to_patch') as delay:
            d = fun_to_patch()
            self.assertEqual(len(delay), 1)
            self.assertFalse(d.called)
            delay.fire()
            self.assertEqual(len(delay), 0)
            self.assertTrue(d.called)

        d = fun_to_patch()
        self.assertTrue(d.called)

    def test_auto_fires_unfired_delay(self):
        with patchForDelay(__name__ + '.fun_to_patch') as delay:
            d = fun_to_patch()
            self.assertEqual(len(delay), 1)
            self.assertFalse(d.called)
        self.assertTrue(d.called)

    def test_auto_fires_unfired_delay_exception(self):
        try:
            with patchForDelay(__name__ + '.fun_to_patch') as delay:
                d = fun_to_patch()
                self.assertEqual(len(delay), 1)
                self.assertFalse(d.called)
                raise TestException()
        except TestException:
            pass
        self.assertTrue(d.called)

    def test_passes_arguments(self):
        with patchForDelay(__name__ + '.fun_to_patch') as delay:
            d = fun_to_patch('arg', kw='kwarg')
            self.assertEqual(len(delay), 1)
            delay.fire()
            args = self.successResultOf(d)

        self.assertEqual(args, (('arg',), {'kw': 'kwarg'}))

    def test_passes_exception(self):
        with patchForDelay(__name__ + '.fun_to_patch_exception') as delay:
            d = fun_to_patch_exception()
            self.assertEqual(len(delay), 1)
            delay.fire()
            f = self.failureResultOf(d)
            f.check(TestException)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

# this is a copy of buildbot.util.Notifier
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import overload

from twisted.internet import defer
from twisted.python import failure
from twisted.python import log

from buildbot_worker.util import Notifier

if TYPE_CHECKING:
    from typing import Any
    from typing import TypeVar

    from buildbot_worker.util.twisted import InlineCallbacksType

    _T = TypeVar('_T')


class DeferWaiter:
    """This class manages a set of Deferred objects and allows waiting for their completion"""

    def __init__(self) -> None:
        self._waited: dict[int, defer.Deferred] = {}
        self._finish_notifier: Notifier[None] = Notifier()

    def _finished(self, result: _T, d: defer.Deferred) -> _T:
        # most likely nothing is consuming the errors, so do it here
        if isinstance(result, failure.Failure):
            log.err(result)

        self._waited.pop(id(d))
        if not self._waited:
            self._finish_notifier.notify(None)
        return result

    @overload
    def add(self, d: defer.Deferred[_T]) -> defer.Deferred[_T]: ...

    @overload
    def add(self, d: Any) -> None: ...

    def add(self, d: defer.Deferred[_T] | Any) -> defer.Deferred[_T] | None:
        if not isinstance(d, defer.Deferred):
            return None

        self._waited[id(d)] = d
        d.addBoth(self._finished, d)
        return d

    def cancel(self) -> None:
        for d in list(self._waited.values()):
            d.cancel()
        self._waited.clear()

    def has_waited(self) -> bool:
        return bool(self._waited)

    @defer.inlineCallbacks
    def wait(self) -> InlineCallbacksType[None]:
        if not self._waited:
            return
        yield self._finish_notifier.wait()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import datetime
import fnmatch
import json
import re
from contextlib import contextmanager
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Iterator
from urllib.parse import urlparse

from twisted.internet import defer
from twisted.internet import threads
from twisted.internet.error import ConnectionDone
from twisted.python import log
from twisted.web.error import Error
from twisted.web.resource import EncodingResourceWrapper
from twisted.web.server import GzipEncoderFactory

from buildbot.data import exceptions
from buildbot.data.base import EndpointKind
from buildbot.data.resultspec import ResultSpec
from buildbot.util import bytes2unicode
from buildbot.util import toJson
from buildbot.util import unicode2bytes
from buildbot.www import resource
from buildbot.www.authz import Forbidden
from buildbot.www.encoding import BrotliEncoderFactory
from buildbot.www.encoding import ZstandardEncoderFactory

if TYPE_CHECKING:
    from typing import Any

    from twisted.web import server

    from buildbot.data.base import Endpoint
    from buildbot.data.resultspec import ResultSpec
    from buildbot.master import BuildMaster
    from buildbot.util.twisted import InlineCallbacksType


class BadJsonRpc2(Exception):
    def __init__(self, message: str, jsonrpccode: int) -> None:
        self.message = message
        self.jsonrpccode = jsonrpccode


class ContentTypeParser:
    def __init__(self, contenttype: str | bytes | None) -> None:
        self.typeheader = contenttype

    def gettype(self) -> str | None:
        if self.typeheader is None:
            return None
        return bytes2unicode(self.typeheader).split(';', 1)[0]


def _is_request_finished(request: server.Request) -> bool:
    # In case of lost connection, request is not marked as finished
    # detect this case with `channel` being None
    return bool(request.finished) or request.channel is None


URL_ENCODED = b"application/x-www-form-urlencoded"
JSON_ENCODED = b"application/json"


class RestRootResource(resource.Resource):
    version_classes: dict[int, type[V2RootResource]] = {}

    @classmethod
    def addApiVersion(cls, version: int, version_cls: type[V2RootResource]) -> None:
        cls.version_classes[version] = version_cls
        version_cls.apiVersion = version  # type: ignore[attr-defined]

    def __init__(self, master: BuildMaster) -> None:
        super().__init__(master)

        min_vers = master.config.www.get('rest_minimum_version', 0)
        encoders = [
            BrotliEncoderFactory(),
            ZstandardEncoderFactory(),
            GzipEncoderFactory(),
        ]

        latest = max(list(self.version_classes))
        for version, klass in self.version_classes.items():
            if version < min_vers:
                continue
            child = EncodingResourceWrapper(klass(master), encoders)
            child_path = f'v{version}'
            self.putChild(unicode2bytes(child_path), child)
            if version == latest:
                self.putChild(b'latest', child)

    def render(self, request: server.Request) -> bytes:
        request.setHeader(b"content-type", JSON_ENCODED)
        min_vers = self.master.config.www.get('rest_minimum_version', 0)
        api_versions = dict(
            (f'v{v}', f'{self.base_url}api/v{v}') for v in self.version_classes if v > min_vers
        )
        data = json.dumps({"api_versions": api_versions})
        return unicode2bytes(data)


JSONRPC_CODES = {
    "parse_error": -32700,
    "invalid_request": -32600,
    "method_not_found": -32601,
    "invalid_params": -32602,
    "internal_error": -32603,
}


class V2RootResource(resource.Resource):
    # For GETs, this API follows http://jsonapi.org.  The getter API does not
    # permit create, update, or delete, so this is limited to reading.
    #
    # Data API control methods can be invoked via a POST to the appropriate
    # URL.  These follow http://www.jsonrpc.org/specification, with a few
    # limitations:
    # - params as list is not supported
    # - rpc call batching is not supported
    # - jsonrpc2 notifications are not supported (you always get an answer)

    # rather than construct the entire possible hierarchy of Rest resources,
    # this is marked as a leaf node, and any remaining path items are parsed
    # during rendering
    isLeaf = True

    # enable reconfigResource calls
    needsReconfig = True

    @defer.inlineCallbacks
    def getEndpoint(
        self, request: server.Request, method: str, params: dict[str, Any]
    ) -> InlineCallbacksType[tuple[Endpoint, dict[str, Any]]]:
        # note that trailing slashes are not allowed
        assert request.postpath is not None
        request_postpath = tuple(bytes2unicode(p) for p in request.postpath)
        yield self.master.www.assertUserAllowed(request, request_postpath, method, params)
        ret = yield self.master.data.getEndpoint(request_postpath)
        return ret

    @contextmanager
    def handleErrors(self, writeError: Callable[[str | bytes, int, int], None]) -> Iterator[None]:
        try:
            yield
        except ConnectionDone:
            # Connection was cleanly closed
            pass
        except exceptions.InvalidPathError as e:
            msg = unicode2bytes(e.args[0])
            writeError(msg or b"invalid path", 404, JSONRPC_CODES['invalid_request'])
            return
        except exceptions.InvalidControlException as e:
            msg = unicode2bytes(str(e))
            writeError(msg or b"invalid control action", 501, JSONRPC_CODES["method_not_found"])
            return
        except exceptions.InvalidQueryParameter as e:
            msg = unicode2bytes(e.args[0])
            writeError(msg or b"invalid request", 400, JSONRPC_CODES["method_not_found"])
            return
        except BadJsonRpc2 as e:
            msg = unicode2bytes(e.message)
            writeError(msg, 400, e.jsonrpccode)
            return
        except Forbidden as e:
            # There is nothing in jsonrc spec about forbidden error, so pick
            # invalid request
            msg = unicode2bytes(e.message)
            writeError(msg, 403, JSONRPC_CODES["invalid_request"])
            return
        except Exception as e:
            log.err(_why='while handling API request')
            msg = unicode2bytes(repr(e))
            writeError(repr(e), 500, JSONRPC_CODES["internal_error"])
            return

    # JSONRPC2 support

    def decodeJsonRPC2(
        self, request: server.Request
    ) -> tuple[str, str | int | None, dict[str, Any]]:
        # Verify the content-type.  Browsers are easily convinced to send
        # POST data to arbitrary URLs via 'form' elements, but they won't
        # use the application/json content-type.
        if ContentTypeParser(request.getHeader(b'content-type')).gettype() != "application/json":
            raise BadJsonRpc2(
                'Invalid content-type (use application/json)', JSONRPC_CODES["invalid_request"]
            )

        try:
            assert request.content is not None
            data = json.loads(bytes2unicode(request.content.read()))
        except Exception as e:
            raise BadJsonRpc2(f"JSON parse error: {e!s}", JSONRPC_CODES["parse_error"]) from e

        if isinstance(data, list):
            raise BadJsonRpc2(
                "JSONRPC batch requests are not supported", JSONRPC_CODES["invalid_request"]
            )
        if not isinstance(data, dict):
            raise BadJsonRpc2(
                "JSONRPC root object must be an object", JSONRPC_CODES["invalid_request"]
            )

        def check(name, types, typename):
            if name not in data:
                raise BadJsonRpc2(f"missing key '{name}'", JSONRPC_CODES["invalid_request"])
            if not isinstance(data[name], types):
                raise BadJsonRpc2(f"'{name}' must be {typename}", JSONRPC_CODES["invalid_request"])

        check("jsonrpc", (str,), "a string")
        check("method", (str,), "a string")
        check("id", (str, int, type(None)), "a string, number, or null")
        check("params", (dict,), "an object")
        if data['jsonrpc'] != '2.0':
            raise BadJsonRpc2("only JSONRPC 2.0 is supported", JSONRPC_CODES['invalid_request'])
        return data["method"], data["id"], data['params']

    @defer.inlineCallbacks
    def renderJsonRpc(self, request: server.Request) -> InlineCallbacksType[None]:
        jsonRpcReply: dict[str, Any] = {'jsonrpc': "2.0"}

        def writeError(
            msg: str | bytes, errcode: int = 399, jsonrpccode: int = JSONRPC_CODES["internal_error"]
        ) -> None:
            if isinstance(msg, bytes):
                msg = bytes2unicode(msg)
            if self.debug:
                log.msg(f"JSONRPC error: {msg}")
            request.setResponseCode(errcode)
            request.setHeader(b'content-type', JSON_ENCODED)
            if "error" not in jsonRpcReply:  # already filled in by caller
                jsonRpcReply['error'] = {"code": jsonrpccode, "message": msg}
            data = json.dumps(jsonRpcReply)
            request.write(unicode2bytes(data))

        with self.handleErrors(writeError):
            method, id, params = self.decodeJsonRPC2(request)
            jsonRpcReply['id'] = id
            ep, kwargs = yield self.getEndpoint(request, method, params)
            userinfos = self.master.www.getUserInfos(request)
            if userinfos.get('anonymous'):
                owner = "anonymous"
            else:
                for field in ('email', 'username', 'full_name'):
                    owner = userinfos.get(field, None)
                    if owner:
                        break
            params['owner'] = owner

            result = yield ep.control(method, params, kwargs)
            jsonRpcReply['result'] = result

            data = json.dumps(jsonRpcReply, default=toJson, sort_keys=True, separators=(',', ':'))

            request.setHeader(b'content-type', JSON_ENCODED)
            if request.method == b"HEAD":
                request.setHeader(b"content-length", unicode2bytes(str(len(data))))
                request.write(b'')
            else:
                request.write(unicode2bytes(data))

    def decodeResultSpec(self, request: server.Request, endpoint: Endpoint) -> ResultSpec:
        args = request.args
        entityType = endpoint.rtype.entityType
        return self.master.data.resultspec_from_jsonapi(
            args, entityType, endpoint.kind == EndpointKind.COLLECTION
        )

    def _write_rest_error(
        self, request: server.Request, msg: str | bytes, errcode: int = 404
    ) -> None:
        if self.debug:
            log.msg(f"REST error: {msg!r}")
        request.setResponseCode(errcode)
        request.setHeader(b'content-type', b'text/plain; charset=utf-8')
        msg = bytes2unicode(msg)
        json_data = json.dumps({"error": msg})
        request.write(unicode2bytes(json_data))

    def _write_not_found_rest_error(
        self,
        request: server.Request,
        ep: Endpoint,
        rspec: ResultSpec,
        kwargs: dict[str, Any],
    ) -> None:
        self._write_rest_error(
            request=request,
            msg=(f"not found while getting from {ep!r} with arguments {rspec!r} and {kwargs!s}"),
        )

    async def _render_raw(
        self,
        request: server.Request,
        ep: Endpoint,
        rspec: ResultSpec,
        kwargs: dict[str, Any],
    ) -> None:
        assert ep.kind in (EndpointKind.RAW, EndpointKind.RAW_INLINE)

        is_stream_data = False
        try:
            data = await ep.stream(rspec, kwargs)
            is_stream_data = True
        except NotImplementedError:
            data = await ep.get(rspec, kwargs)

        if data is None:
            self._write_not_found_rest_error(request, ep, rspec=rspec, kwargs=kwargs)
            return

        request.setHeader(b"content-type", unicode2bytes(data['mime-type']) + b'; charset=utf-8')
        if ep.kind != EndpointKind.RAW_INLINE:
            request.setHeader(
                b"content-disposition", b'attachment; filename=' + unicode2bytes(data['filename'])
            )

        if not is_stream_data:
            request.write(unicode2bytes(data['raw']))
            return

        async for chunk in data['raw']:
            if _is_request_finished(request):
                return
            request.write(unicode2bytes(chunk))

    @defer.inlineCallbacks
    def renderRest(self, request: server.Request) -> InlineCallbacksType[None]:
        def writeError(
            msg: str | bytes, errcode: int = 404, jsonrpccode: int | None = None
        ) -> None:
            self._write_rest_error(request, msg=msg, errcode=errcode)

        with self.handleErrors(writeError):
            ep, kwargs = yield self.getEndpoint(request, bytes2unicode(request.method), {})

            rspec = self.decodeResultSpec(request, ep)
            if ep.kind in (EndpointKind.RAW, EndpointKind.RAW_INLINE):
                yield defer.Deferred.fromCoroutine(self._render_raw(request, ep, rspec, kwargs))
                return

            data = yield ep.get(rspec, kwargs)
            if data is None:
                self._write_not_found_rest_error(request, ep, rspec=rspec, kwargs=kwargs)
                return

            if _is_request_finished(request):
                return

            # post-process any remaining parts of the resultspec
            data = rspec.apply(data)

            # annotate the result with some metadata
            meta = {}
            if ep.kind == EndpointKind.COLLECTION:
                offset, total = data.offset, data.total
                if offset is None:
                    offset = 0

                # add total, if known
                if total is not None:
                    meta['total'] = total

                # get the real list instance out of the ListResult
                data = data.data
            else:
                data = [data]

            typeName = ep.rtype.plural
            data = {typeName: data, 'meta': meta}

            # set up the content type and formatting options; if the request
            # accepts text/html or text/plain, the JSON will be rendered in a
            # readable, multiline format.

            if b'application/json' in (request.getHeader(b'accept') or b''):
                compact = True
                request.setHeader(b"content-type", b'application/json; charset=utf-8')
            else:
                compact = False
                request.setHeader(b"content-type", b'text/plain; charset=utf-8')

            # set up caching
            if self.cache_seconds:
                now = datetime.datetime.now(datetime.timezone.utc)
                expires = now + datetime.timedelta(seconds=self.cache_seconds)
                expiresBytes = unicode2bytes(expires.strftime("%a, %d %b %Y %H:%M:%S GMT"))
                request.setHeader(b"Expires", expiresBytes)
                request.setHeader(b"Pragma", b"no-cache")

            # filter out blanks if necessary and render the data
            encoder = json.encoder.JSONEncoder(default=toJson, sort_keys=True)
            if compact:
                encoder.item_separator, encoder.key_separator = (',', ':')
            else:
                encoder.indent = 2

            yield threads.deferToThread(V2RootResource._write_json_data, request, encoder, data)

    def reconfigResource(self, new_config: Any) -> None:
        # buildbotURL may contain reverse proxy path, Origin header is just
        # scheme + host + port
        buildbotURL = urlparse(unicode2bytes(new_config.buildbotURL))
        origin_self = buildbotURL.scheme + b"://" + buildbotURL.netloc
        # pre-translate the origin entries in the config
        self.origins = []
        for o in new_config.www.get('allowed_origins', [origin_self]):
            origin = bytes2unicode(o).lower()
            self.origins.append(re.compile(fnmatch.translate(origin)))

        # and copy some other flags
        self.debug = new_config.www.get('debug')
        self.cache_seconds = new_config.www.get('json_cache_seconds', 0)

    def render(self, request: server.Request) -> int:
        def writeError(msg: str | bytes, errcode: int = 400) -> None:
            msg = bytes2unicode(msg)
            if self.debug:
                log.msg(f"HTTP error: {msg}")
            request.setResponseCode(errcode)
            request.setHeader(b'content-type', b'text/plain; charset=utf-8')
            if request.method == b'POST':
                # jsonRPC callers want the error message in error.message
                data = json.dumps({"error": {"message": msg}})
                request.write(unicode2bytes(data))
            else:
                data = json.dumps({"error": msg})
                request.write(unicode2bytes(data))
            request.finish()

        return self.asyncRenderHelper(request, self.asyncRender, writeError)

    @defer.inlineCallbacks
    def asyncRender(self, request: server.Request) -> InlineCallbacksType[bytes | None]:
        # Handle CORS, if necessary.
        origins = self.origins
        if origins is not None:
            isPreflight = False
            reqOrigin = request.getHeader(b'origin')
            if reqOrigin:
                err = None
                reqOrigin = reqOrigin.lower()
                if not any(o.match(bytes2unicode(reqOrigin)) for o in self.origins):
                    err = b"invalid origin"
                elif request.method == b'OPTIONS':
                    preflightMethod = request.getHeader(b'access-control-request-method')
                    if preflightMethod not in (b'GET', b'POST', b'HEAD'):
                        err = b'invalid method'
                    isPreflight = True
                if err:
                    raise Error(400, err)

                # If it's OK, then let the browser know we checked it out.  The
                # Content-Type header is included here because CORS considers
                # content types other than form data and text/plain to not be
                # simple.
                request.setHeader(b"access-control-allow-origin", reqOrigin)
                request.setHeader(b"access-control-allow-headers", b"Content-Type")
                request.setHeader(b"access-control-max-age", b'3600')

                # if this was a preflight request, we're done
                if isPreflight:
                    return b""

        # based on the method, this is either JSONRPC or REST
        if request.method == b'POST':
            res = yield self.renderJsonRpc(request)
        elif request.method in (b'GET', b'HEAD'):
            res = yield self.renderRest(request)
        else:
            raise Error(400, b"invalid HTTP method")

        return res

    @staticmethod
    def _write_json_data(
        request: server.Request,
        encoder: json.encoder.JSONEncoder,
        data: Any,
    ) -> None:
        content_length = 0
        for chunk in encoder.iterencode(data):
            if _is_request_finished(request):
                return
            content_length += len(unicode2bytes(chunk))
        request.setHeader(b"content-length", unicode2bytes(str(content_length)))

        if request.method != b"HEAD":
            for chunk in encoder.iterencode(data):
                if _is_request_finished(request):
                    return
                request.write(unicode2bytes(chunk))


RestRootResource.addApiVersion(2, V2RootResource)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import os
import platform
import textwrap
from typing import Any

from twisted.trial import unittest

from buildbot.config import BuilderConfig
from buildbot.db import enginestrategy
from buildbot.plugins import schedulers
from buildbot.process.factory import BuildFactory
from buildbot.scripts import copydb
from buildbot.steps.master import MasterShellCommand
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import db
from buildbot.test.util import dirs
from buildbot.test.util import misc
from buildbot.test.util.integration import RunMasterBase
from buildbot.util.twisted import async_to_deferred


def get_script_config(destination_url: str = 'sqlite://', **kwargs: object) -> dict[str, Any]:
    config = {
        "quiet": False,
        "basedir": os.path.abspath('basedir'),
        'destination_url': destination_url,
        'ignore-fk-error-rows': True,
    }
    config.update(kwargs)
    return config


def write_buildbot_tac(path: str) -> None:
    with open(path, "w", encoding='utf-8') as f:
        f.write(
            textwrap.dedent("""
            from twisted.application import service
            application = service.Application('buildmaster')
        """)
        )


class TestCopyDb(misc.StdoutAssertionsMixin, dirs.DirsMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setup_test_reactor()
        self.setUpDirs('basedir')
        write_buildbot_tac(os.path.join('basedir', 'buildbot.tac'))
        self.setUpStdoutAssertions()

    def create_master_cfg(self, db_url: str = 'sqlite://', extraconfig: str = "") -> None:
        with open(os.path.join('basedir', 'master.cfg'), "w", encoding='utf-8') as f:
            f.write(
                textwrap.dedent(f"""
                from buildbot.plugins import *
                c = BuildmasterConfig = dict()
                c['db_url'] = {db_url!r}
                c['buildbotNetUsageData'] = None
                c['multiMaster'] = True  # don't complain for no builders
                {extraconfig}
            """)
            )

    @async_to_deferred
    async def test_not_basedir(self) -> None:
        res = await copydb._copy_database_in_reactor(get_script_config(basedir='doesntexist'))
        self.assertEqual(res, 1)
        tac_path = os.path.join('doesntexist', 'buildbot.tac')
        self.assertInStdout(f'error reading \'{tac_path}\': No such file or directory')

    @async_to_deferred
    async def test_bad_config(self) -> None:
        res = await copydb._copy_database_in_reactor(get_script_config(basedir='basedir'))
        self.assertEqual(res, 1)
        master_cfg_path = os.path.abspath(os.path.join('basedir', 'master.cfg'))
        self.assertInStdout(f'configuration file \'{master_cfg_path}\' does not exist')

    @async_to_deferred
    async def test_bad_config2(self) -> None:
        self.create_master_cfg(extraconfig="++++ # syntaxerror")
        res = await copydb._copy_database_in_reactor(get_script_config(basedir='basedir'))
        self.assertEqual(res, 1)
        self.assertInStdout("encountered a SyntaxError while parsing config file:")
        # config logs an error via log.err, we must eat it or trial will
        # complain
        self.flushLoggedErrors()


class TestCopyDbRealDb(misc.StdoutAssertionsMixin, RunMasterBase, dirs.DirsMixin, TestReactorMixin):
    INITIAL_DB_URL = 'sqlite:///tmp.sqlite'

    def setUp(self) -> None:
        self.setUpDirs('basedir')
        self.setUpStdoutAssertions()  # comment out to see stdout from script
        write_buildbot_tac(os.path.join('basedir', 'buildbot.tac'))

    async def create_master_config(self) -> int:
        f = BuildFactory()
        cmd = "dir" if platform.system() in ("Windows", "Microsoft") else "ls"
        f.addStep(MasterShellCommand(cmd))

        config_dict = {
            'builders': [
                BuilderConfig(
                    name="testy",
                    workernames=["local1"],
                    factory=f,
                ),
            ],
            'schedulers': [schedulers.ForceScheduler(name="force", builderNames=["testy"])],
            # Disable checks about missing scheduler.
            'multiMaster': True,
            'db_url': self.INITIAL_DB_URL,
        }
        await self.setup_master(config_dict, basedir='basedir')
        builder_id = await self.master.data.updates.findBuilderId('testy')

        return builder_id

    def create_master_config_file(self, db_url: str) -> None:
        with open(os.path.join('basedir', 'master.cfg'), "w", encoding='utf-8') as f:
            f.write(
                textwrap.dedent(f"""
                from buildbot.plugins import *
                c = BuildmasterConfig = dict()
                c['db_url'] = {db_url!r}
                c['buildbotNetUsageData'] = None
                c['multiMaster'] = True  # don't complain for no builders
            """)
            )

    def resolve_db_url(self) -> str:
        # test may use mysql or pg if configured in env
        envkey = "BUILDBOT_TEST_DB_URL"
        if envkey not in os.environ or os.environ[envkey] == 'sqlite://':
            return "sqlite:///" + os.path.abspath(os.path.join("basedir", "dest.sqlite"))
        return os.environ[envkey]

    def drop_database_tables(self, db_url: str) -> None:
        engine = enginestrategy.create_engine(db_url, basedir='basedir')
        with engine.connect() as conn:
            db.thd_clean_database(conn)
        engine.dispose()

    @async_to_deferred
    async def test_full(self) -> None:
        await self.create_master_config()

        await self.doForceBuild()
        await self.doForceBuild()
        await self.doForceBuild()

        await self.tested_master.shutdown()

        self.create_master_config_file(self.INITIAL_DB_URL)

        dest_db_url = db.resolve_test_index_in_db_url(self.resolve_db_url())

        self.drop_database_tables(dest_db_url)
        self.addCleanup(lambda: self.drop_database_tables(dest_db_url))

        script_config = get_script_config(destination_url=dest_db_url)
        res = await copydb._copy_database_in_reactor(script_config)
        self.assertEqual(res, 0)



import json

from klein import Klein
from twisted.internet import defer


class Api:
    app = Klein()
    pizzaIngredients = {
        'margherita': ['tomato', 'ham', 'cheese'],
        'regina': ['tomato', 'ham', 'cheese', 'mushrooms'],
    }

    def __init__(self, ep):
        self.ep = ep

    @app.route("/getIngredients", methods=['GET'])
    def getIngredients(self, request):
        pizzaArgument = request.args.get('pizza')
        if pizzaArgument is None:
            return defer.succeed(json.dumps("invalid request"))
        pizza = pizzaArgument[0].lower()
        res = self.pizzaIngredients.get(
            pizza, [f"only {self.pizzaIngredients.keys()} are supported for now"]
        )
        return defer.succeed(json.dumps(res))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os
from unittest import mock

from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.machine.generic import HttpAction
from buildbot.machine.generic import LocalWakeAction
from buildbot.machine.generic import LocalWOLAction
from buildbot.machine.generic import RemoteSshSuspendAction
from buildbot.machine.generic import RemoteSshWakeAction
from buildbot.machine.generic import RemoteSshWOLAction
from buildbot.test.fake import fakemaster
from buildbot.test.fake import httpclientservice as fakehttpclientservice
from buildbot.test.fake.private_tempdir import MockPrivateTemporaryDirectory
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.runprocess import ExpectMasterShell
from buildbot.test.runprocess import MasterRunProcessMixin
from buildbot.test.util import config


class FakeManager:
    def __init__(self, master):
        self.master = master

    def renderSecrets(self, args):
        return defer.succeed(args)


def create_simple_mock_master(reactor, basedir=None):
    master = mock.Mock()
    master.basedir = basedir
    master.reactor = reactor
    return master


class TestActions(
    MasterRunProcessMixin, config.ConfigErrorsMixin, TestReactorMixin, unittest.TestCase
):
    def setUp(self):
        self.setup_test_reactor()
        self.setup_master_run_process()

    @defer.inlineCallbacks
    def test_local_wake_action(self):
        self.expect_commands(
            ExpectMasterShell(['cmd', 'arg1', 'arg2']).exit(1),
            ExpectMasterShell(['cmd', 'arg1', 'arg2']).exit(0),
        )

        manager = FakeManager(create_simple_mock_master(self.reactor))
        action = LocalWakeAction(['cmd', 'arg1', 'arg2'])
        self.assertFalse((yield action.perform(manager)))
        self.assertTrue((yield action.perform(manager)))
        self.assert_all_commands_ran()

    def test_local_wake_action_command_not_list(self):
        with self.assertRaisesConfigError('command parameter must be a list'):
            LocalWakeAction('not-list')

    @defer.inlineCallbacks
    def test_local_wol_action(self):
        self.expect_commands(
            ExpectMasterShell(['wol', '00:11:22:33:44:55']).exit(1),
            ExpectMasterShell(['wakeonlan', '00:11:22:33:44:55']).exit(0),
        )

        manager = FakeManager(create_simple_mock_master(self.reactor))
        action = LocalWOLAction('00:11:22:33:44:55', wolBin='wol')
        self.assertFalse((yield action.perform(manager)))

        action = LocalWOLAction('00:11:22:33:44:55')
        self.assertTrue((yield action.perform(manager)))
        self.assert_all_commands_ran()

    @mock.patch(
        'buildbot.util.private_tempdir.PrivateTemporaryDirectory',
        new_callable=MockPrivateTemporaryDirectory,
    )
    @mock.patch('buildbot.util.misc.writeLocalFile')
    @defer.inlineCallbacks
    def test_remote_ssh_wake_action_no_keys(self, write_local_file_mock, temp_dir_mock):
        self.expect_commands(
            ExpectMasterShell([
                'ssh',
                '-o',
                'BatchMode=yes',
                'remote_host',
                'remotebin',
                'arg1',
            ]).exit(1),
            ExpectMasterShell([
                'ssh',
                '-o',
                'BatchMode=yes',
                'remote_host',
                'remotebin',
                'arg1',
            ]).exit(0),
        )

        manager = FakeManager(create_simple_mock_master(self.reactor))
        action = RemoteSshWakeAction('remote_host', ['remotebin', 'arg1'])
        self.assertFalse((yield action.perform(manager)))
        self.assertTrue((yield action.perform(manager)))
        self.assert_all_commands_ran()

        self.assertEqual(temp_dir_mock.dirs, [])
        write_local_file_mock.assert_not_called()

    @mock.patch(
        'buildbot.util.private_tempdir.PrivateTemporaryDirectory',
        new_callable=MockPrivateTemporaryDirectory,
    )
    @mock.patch('buildbot.util.misc.writeLocalFile')
    @defer.inlineCallbacks
    def test_remote_ssh_wake_action_with_keys(self, write_local_file_mock, temp_dir_mock):
        temp_dir_path = os.path.join('path-to-master', 'ssh-@@@')
        ssh_key_path = os.path.join(temp_dir_path, 'ssh-key')
        ssh_known_hosts_path = os.path.join(temp_dir_path, 'ssh-known-hosts')

        self.expect_commands(
            ExpectMasterShell([
                'ssh',
                '-o',
                'BatchMode=yes',
                '-i',
                ssh_key_path,
                '-o',
                f'UserKnownHostsFile={ssh_known_hosts_path}',
                'remote_host',
                'remotebin',
                'arg1',
            ]).exit(0),
        )

        manager = FakeManager(create_simple_mock_master(self.reactor, 'path-to-master'))
        action = RemoteSshWakeAction(
            'remote_host', ['remotebin', 'arg1'], sshKey='ssh_key', sshHostKey='ssh_host_key'
        )
        self.assertTrue((yield action.perform(manager)))

        self.assert_all_commands_ran()

        self.assertEqual(temp_dir_mock.dirs, [(temp_dir_path, 0o700)])

        self.assertSequenceEqual(
            write_local_file_mock.call_args_list,
            [
                mock.call(ssh_key_path, 'ssh_key', mode=0o400),
                mock.call(ssh_known_hosts_path, '* ssh_host_key'),
            ],
        )

    def test_remote_ssh_wake_action_sshBin_not_str(self):
        with self.assertRaisesConfigError('sshBin parameter must be a string'):
            RemoteSshWakeAction('host', ['cmd'], sshBin=123)

    def test_remote_ssh_wake_action_host_not_str(self):
        with self.assertRaisesConfigError('host parameter must be a string'):
            RemoteSshWakeAction(123, ['cmd'])

    def test_remote_ssh_wake_action_command_not_list(self):
        with self.assertRaisesConfigError('remoteCommand parameter must be a list'):
            RemoteSshWakeAction('host', 'cmd')

    @mock.patch(
        'buildbot.util.private_tempdir.PrivateTemporaryDirectory',
        new_callable=MockPrivateTemporaryDirectory,
    )
    @mock.patch('buildbot.util.misc.writeLocalFile')
    @defer.inlineCallbacks
    def test_remote_ssh_wol_action_no_keys(self, write_local_file_mock, temp_dir_mock):
        self.expect_commands(
            ExpectMasterShell([
                'ssh',
                '-o',
                'BatchMode=yes',
                'remote_host',
                'wakeonlan',
                '00:11:22:33:44:55',
            ]).exit(0),
            ExpectMasterShell([
                'ssh',
                '-o',
                'BatchMode=yes',
                'remote_host',
                'wolbin',
                '00:11:22:33:44:55',
            ]).exit(0),
        )

        manager = FakeManager(create_simple_mock_master(self.reactor))
        action = RemoteSshWOLAction('remote_host', '00:11:22:33:44:55')
        self.assertTrue((yield action.perform(manager)))

        action = RemoteSshWOLAction('remote_host', '00:11:22:33:44:55', wolBin='wolbin')
        self.assertTrue((yield action.perform(manager)))
        self.assert_all_commands_ran()

        self.assertEqual(temp_dir_mock.dirs, [])
        write_local_file_mock.assert_not_called()

    @mock.patch(
        'buildbot.util.private_tempdir.PrivateTemporaryDirectory',
        new_callable=MockPrivateTemporaryDirectory,
    )
    @mock.patch('buildbot.util.misc.writeLocalFile')
    @defer.inlineCallbacks
    def test_remote_ssh_suspend_action_no_keys(self, write_local_file_mock, temp_dir_mock):
        self.expect_commands(
            ExpectMasterShell([
                'ssh',
                '-o',
                'BatchMode=yes',
                'remote_host',
                'systemctl',
                'suspend',
            ]).exit(0),
            ExpectMasterShell([
                'ssh',
                '-o',
                'BatchMode=yes',
                'remote_host',
                'dosuspend',
                'arg1',
            ]).exit(0),
        )

        manager = FakeManager(create_simple_mock_master(self.reactor))
        action = RemoteSshSuspendAction('remote_host')
        self.assertTrue((yield action.perform(manager)))

        action = RemoteSshSuspendAction('remote_host', remoteCommand=['dosuspend', 'arg1'])
        self.assertTrue((yield action.perform(manager)))
        self.assert_all_commands_ran()

        self.assertEqual(temp_dir_mock.dirs, [])
        write_local_file_mock.assert_not_called()


class TestHttpAction(config.ConfigErrorsMixin, TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)
        self.http = yield fakehttpclientservice.HTTPClientService.getService(
            self.master, self, "http://localhost/request"
        )

    @defer.inlineCallbacks
    def test_http_wrong_method(self):
        manager = FakeManager(self.master)
        action = HttpAction('http://localhost/request', method='non-existing-method')

        with self.assertRaisesConfigError('Invalid method non-existing-method'):
            yield action.perform(manager)

    @parameterized.expand([
        'get',
        'post',
        'delete',
        'put',
    ])
    @defer.inlineCallbacks
    def test_http(self, method):
        self.http.expect(method, '')

        manager = FakeManager(self.master)
        action = HttpAction('http://localhost/request', method=method)

        yield action.perform(manager)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from twisted.internet import defer
from twisted.internet import reactor
from twisted.spread import pb
from twisted.trial import unittest

from buildbot.clients import sendchange


class Sender(unittest.TestCase):
    def setUp(self):
        # patch out some PB components and make up some mocks
        self.patch(pb, 'PBClientFactory', self._fake_PBClientFactory)
        self.patch(reactor, 'connectTCP', self._fake_connectTCP)

        self.factory = mock.Mock(name='PBClientFactory')
        self.factory.login = self._fake_login
        self.factory.login_d = defer.Deferred()

        self.remote = mock.Mock(name='PB Remote')
        self.remote.callRemote = self._fake_callRemote
        self.remote.broker.transport.loseConnection = self._fake_loseConnection

        # results
        self.creds = None
        self.conn_host = self.conn_port = None
        self.lostConnection = False
        self.added_changes = []
        self.vc_used = None

    def _fake_PBClientFactory(self):
        return self.factory

    def _fake_login(self, creds):
        self.creds = creds
        return self.factory.login_d

    def _fake_connectTCP(self, host, port, factory):
        self.conn_host = host
        self.conn_port = port
        self.assertIdentical(factory, self.factory)
        self.factory.login_d.callback(self.remote)

    def _fake_callRemote(self, method, change):
        self.assertEqual(method, 'addChange')
        self.added_changes.append(change)
        return defer.succeed(None)

    def _fake_loseConnection(self):
        self.lostConnection = True

    def assertProcess(self, host, port, username, password, changes):
        self.assertEqual(
            [host, port, username, password, changes],
            [
                self.conn_host,
                self.conn_port,
                self.creds.username,
                self.creds.password,
                self.added_changes,
            ],
        )

    @defer.inlineCallbacks
    def test_send_minimal(self):
        s = sendchange.Sender('localhost:1234')
        yield s.send('branch', 'rev', 'comm', ['a'])

        self.assertProcess(
            'localhost',
            1234,
            b'change',
            b'changepw',
            [
                {
                    "project": '',
                    "repository": '',
                    "who": None,
                    "files": ['a'],
                    "comments": 'comm',
                    "branch": 'branch',
                    "revision": 'rev',
                    "category": None,
                    "when": None,
                    "properties": {},
                    "revlink": '',
                    "src": None,
                }
            ],
        )

    @defer.inlineCallbacks
    def test_send_auth(self):
        s = sendchange.Sender('localhost:1234', auth=('me', 'sekrit'))
        yield s.send('branch', 'rev', 'comm', ['a'])

        self.assertProcess(
            'localhost',
            1234,
            b'me',
            b'sekrit',
            [
                {
                    "project": '',
                    "repository": '',
                    "who": None,
                    "files": ['a'],
                    "comments": 'comm',
                    "branch": 'branch',
                    "revision": 'rev',
                    "category": None,
                    "when": None,
                    "properties": {},
                    "revlink": '',
                    "src": None,
                }
            ],
        )

    @defer.inlineCallbacks
    def test_send_full(self):
        s = sendchange.Sender('localhost:1234')
        yield s.send(
            'branch',
            'rev',
            'comm',
            ['a'],
            who='me',
            category='cats',
            when=1234,
            properties={'a': 'b'},
            repository='r',
            vc='git',
            project='p',
            revlink='rl',
        )

        self.assertProcess(
            'localhost',
            1234,
            b'change',
            b'changepw',
            [
                {
                    "project": 'p',
                    "repository": 'r',
                    "who": 'me',
                    "files": ['a'],
                    "comments": 'comm',
                    "branch": 'branch',
                    "revision": 'rev',
                    "category": 'cats',
                    "when": 1234,
                    "properties": {'a': 'b'},
                    "revlink": 'rl',
                    "src": 'git',
                }
            ],
        )

    @defer.inlineCallbacks
    def test_send_files_tuple(self):
        # 'buildbot sendchange' sends files as a tuple, rather than a list..
        s = sendchange.Sender('localhost:1234')
        yield s.send('branch', 'rev', 'comm', ('a', 'b'))

        self.assertProcess(
            'localhost',
            1234,
            b'change',
            b'changepw',
            [
                {
                    "project": '',
                    "repository": '',
                    "who": None,
                    "files": ['a', 'b'],
                    "comments": 'comm',
                    "branch": 'branch',
                    "revision": 'rev',
                    "category": None,
                    "when": None,
                    "properties": {},
                    "revlink": '',
                    "src": None,
                }
            ],
        )

    @defer.inlineCallbacks
    def test_send_codebase(self):
        s = sendchange.Sender('localhost:1234')
        yield s.send('branch', 'rev', 'comm', ['a'], codebase='mycb')

        self.assertProcess(
            'localhost',
            1234,
            b'change',
            b'changepw',
            [
                {
                    "project": '',
                    "repository": '',
                    "who": None,
                    "files": ['a'],
                    "comments": 'comm',
                    "branch": 'branch',
                    "revision": 'rev',
                    "category": None,
                    "when": None,
                    "properties": {},
                    "revlink": '',
                    "src": None,
                    "codebase": 'mycb',
                }
            ],
        )

    @defer.inlineCallbacks
    def test_send_unicode(self):
        s = sendchange.Sender('localhost:1234')
        yield s.send(
            '\N{DEGREE SIGN}',
            '\U0001f49e',
            '\N{POSTAL MARK FACE}',
            ['\U0001f4c1'],
            project='\N{SKULL AND CROSSBONES}',
            repository='\N{SNOWMAN}',
            who='\N{THAI CHARACTER KHOMUT}',
            category='\U0001f640',
            when=1234,
            properties={'\N{LATIN SMALL LETTER A WITH MACRON}': 'b'},
            revlink='\U0001f517',
        )

        self.assertProcess(
            'localhost',
            1234,
            b'change',
            b'changepw',
            [
                {
                    "project": '\N{SKULL AND CROSSBONES}',
                    "repository": '\N{SNOWMAN}',
                    "who": '\N{THAI CHARACTER KHOMUT}',
                    "files": ['\U0001f4c1'],  # FILE FOLDER
                    "comments": '\N{POSTAL MARK FACE}',
                    "branch": '\N{DEGREE SIGN}',
                    "revision": '\U0001f49e',  # REVOLVING HEARTS
                    "category": '\U0001f640',  # WEARY CAT FACE
                    "when": 1234,
                    "properties": {'\N{LATIN SMALL LETTER A WITH MACRON}': 'b'},
                    "revlink": '\U0001f517',  # LINK SYMBOL
                    "src": None,
                }
            ],
        )

    @defer.inlineCallbacks
    def test_send_unicode_utf8(self):
        s = sendchange.Sender('localhost:1234')

        yield s.send(
            '\N{DEGREE SIGN}'.encode(),
            '\U0001f49e'.encode(),
            '\N{POSTAL MARK FACE}'.encode(),
            ['\U0001f4c1'.encode()],
            project='\N{SKULL AND CROSSBONES}'.encode(),
            repository='\N{SNOWMAN}'.encode(),
            who='\N{THAI CHARACTER KHOMUT}'.encode(),
            category='\U0001f640'.encode(),
            when=1234,
            properties={'\N{LATIN SMALL LETTER A WITH MACRON}'.encode(): 'b'},
            revlink='\U0001f517'.encode(),
        )

        self.assertProcess(
            'localhost',
            1234,
            b'change',
            b'changepw',
            [
                {
                    "project": '\N{SKULL AND CROSSBONES}',
                    "repository": '\N{SNOWMAN}',
                    "who": '\N{THAI CHARACTER KHOMUT}',
                    "files": ['\U0001f4c1'],  # FILE FOLDER
                    "comments": '\N{POSTAL MARK FACE}',
                    "branch": '\N{DEGREE SIGN}',
                    "revision": '\U0001f49e',  # REVOLVING HEARTS
                    "category": '\U0001f640',  # WEARY CAT FACE
                    "when": 1234,
                    # NOTE: not decoded!
                    "properties": {b'\xc4\x81': 'b'},
                    "revlink": '\U0001f517',  # LINK SYMBOL
                    "src": None,
                }
            ],
        )

    @defer.inlineCallbacks
    def test_send_unicode_latin1(self):
        # hand send() a bunch of latin1 strings, and expect them recoded
        # to unicode
        s = sendchange.Sender('localhost:1234', encoding='latin1')

        yield s.send(
            '\N{YEN SIGN}'.encode('latin1'),
            '\N{POUND SIGN}'.encode('latin1'),
            '\N{BROKEN BAR}'.encode('latin1'),
            ['\N{NOT SIGN}'.encode('latin1')],
            project='\N{DEGREE SIGN}'.encode('latin1'),
            repository='\N{SECTION SIGN}'.encode('latin1'),
            who='\N{MACRON}'.encode('latin1'),
            category='\N{PILCROW SIGN}'.encode('latin1'),
            when=1234,
            properties={'\N{SUPERSCRIPT ONE}'.encode('latin1'): 'b'},
            revlink='\N{INVERTED QUESTION MARK}'.encode('latin1'),
        )

        self.assertProcess(
            'localhost',
            1234,
            b'change',
            b'changepw',
            [
                {
                    "project": '\N{DEGREE SIGN}',
                    "repository": '\N{SECTION SIGN}',
                    "who": '\N{MACRON}',
                    "files": ['\N{NOT SIGN}'],
                    "comments": '\N{BROKEN BAR}',
                    "branch": '\N{YEN SIGN}',
                    "revision": '\N{POUND SIGN}',
                    "category": '\N{PILCROW SIGN}',
                    "when": 1234,
                    # NOTE: not decoded!
                    "properties": {b'\xb9': 'b'},
                    "revlink": '\N{INVERTED QUESTION MARK}',
                    "src": None,
                }
            ],
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from twisted.internet import defer
from twisted.internet import reactor
from twisted.trial import unittest
from twisted.web.resource import Resource
from twisted.web.util import redirectTo

from buildbot.process import properties
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.steps import http
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import TestBuildStepMixin
from buildbot.test.util.site import SiteWithClose

try:
    import txrequests

    assert txrequests
    import requests

    assert requests
except ImportError:
    txrequests = requests = None


# We use twisted's internal webserver instead of mocking requests
# to be sure we use the correct requests interfaces


class TestPage(Resource):
    isLeaf = True

    def render_GET(self, request):
        if request.uri == b"/404":
            request.setResponseCode(404)
            return b"404"
        elif request.uri == b'/redirect':
            return redirectTo(b'/redirected-path', request)
        elif request.uri == b"/header":
            return b"".join(request.requestHeaders.getRawHeaders(b"X-Test"))
        return b"OK"

    def render_POST(self, request):
        if request.uri == b"/404":
            request.setResponseCode(404)
            return b"404"
        return b"OK:" + request.content.read()


class TestHTTPStep(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        if txrequests is None:
            raise unittest.SkipTest("Need to install txrequests to test http steps")

        # ignore 'http_proxy' environment variable when running tests
        session = http.getSession()
        self.addCleanup(http.closeSession)
        session.trust_env = False

        # port 0 means random unused port
        self.site = SiteWithClose(TestPage())
        self.addCleanup(self.site.stopFactory)
        self.addCleanup(self.site.close_connections)

        self.listener = reactor.listenTCP(0, self.site)
        self.addCleanup(self.listener.stopListening)

        self.port = self.listener.getHost().port
        return self.setup_test_build_step()

    def get_connection_string(self):
        return f"http://127.0.0.1:{self.port}"

    def getURL(self, path=""):
        return f'{self.get_connection_string()}/{path}'

    def test_get(self):
        url = self.getURL()
        self.setup_step(http.GET(url))
        self.expect_log_file('log', f"URL: {url}\nStatus: 200\n ------ Content ------\nOK")
        self.expect_log_file('content', "OK")
        self.expect_outcome(result=SUCCESS, state_string="Status code: 200")
        return self.run_step()

    def test_connection_error(self):
        def throwing_request(*args, **kwargs):
            raise requests.exceptions.ConnectionError("failed to connect")

        with mock.patch.object(http.getSession(), 'request', throwing_request):
            url = self.getURL("path")
            self.setup_step(http.GET(url))
            self.expect_outcome(result=FAILURE, state_string="Requested (failure)")
            return self.run_step()

    def test_redirect(self):
        url = self.getURL("redirect")
        self.setup_step(http.GET(url))

        expected_log = f"""
Redirected 1 times:

URL: {self.get_connection_string()}/redirect
 ------ Content ------

<html>
    <head>
        <meta http-equiv="refresh" content="0;URL=/redirected-path">
    </head>
    <body bgcolor="#FFFFFF" text="#000000">
    <a href="/redirected-path">click here</a>
    </body>
</html>
============================================================
URL: {self.get_connection_string()}/redirected-path
Status: 200
 ------ Content ------
OK"""

        self.expect_log_file('log', expected_log)
        self.expect_log_file('content', "OK")
        self.expect_outcome(result=SUCCESS, state_string="Status code: 200")
        return self.run_step()

    def test_404(self):
        url = self.getURL("404")
        self.setup_step(http.GET(url))
        self.expect_log_file('log', f"URL: {url}\n ------ Content ------\n404")
        self.expect_log_file('content', "404")
        self.expect_outcome(result=FAILURE, state_string="Status code: 404 (failure)")
        return self.run_step()

    def test_method_not_allowed(self):
        url = self.getURL("path")
        self.setup_step(http.PUT(url))
        self.expect_outcome(result=FAILURE, state_string="Status code: 501 (failure)")
        return self.run_step()

    def test_post(self):
        url = self.getURL("path")
        self.setup_step(http.POST(url))
        self.expect_outcome(result=SUCCESS, state_string="Status code: 200")
        self.expect_log_file('log', f"URL: {url}\nStatus: 200\n ------ Content ------\nOK:")
        self.expect_log_file('content', "OK:")
        return self.run_step()

    def test_post_data(self):
        url = self.getURL("path")
        self.setup_step(http.POST(url, data='mydata'))
        self.expect_outcome(result=SUCCESS, state_string="Status code: 200")
        self.expect_log_file('log', f"URL: {url}\nStatus: 200\n ------ Content ------\nOK:mydata")
        self.expect_log_file('content', "OK:mydata")
        return self.run_step()

    def test_post_data_dict(self):
        url = self.getURL("path")

        self.setup_step(http.POST(url, data={'key1': 'value1'}))
        self.expect_outcome(result=SUCCESS, state_string="Status code: 200")
        self.expect_log_file(
            'log',
            f"""\
URL: {url}
Status: 200
 ------ Content ------
OK:key1=value1""",
        )
        self.expect_log_file('content', "OK:key1=value1")
        return self.run_step()

    def test_header(self):
        url = self.getURL("header")
        self.setup_step(http.GET(url, headers={"X-Test": "True"}))
        self.expect_log_file('log', f"URL: {url}\nStatus: 200\n ------ Content ------\nTrue")
        self.expect_outcome(result=SUCCESS, state_string="Status code: 200")
        return self.run_step()

    @defer.inlineCallbacks
    def test_hidden_header(self):
        url = self.getURL("header")
        self.setup_step(
            http.GET(
                url,
                headers={"X-Test": "True"},
                hide_request_headers=["X-Test"],
                hide_response_headers=["Content-Length"],
            )
        )
        self.expect_log_file('log', f"URL: {url}\nStatus: 200\n ------ Content ------\nTrue")
        self.expect_outcome(result=SUCCESS, state_string="Status code: 200")
        yield self.run_step()
        self.assertIn("X-Test: <HIDDEN>", self.get_nth_step(0).logs['log'].header)
        self.assertIn("Content-Length: <HIDDEN>", self.get_nth_step(0).logs['log'].header)

    def test_params_renderable(self):
        url = self.getURL()
        self.setup_step(http.GET(url, params=properties.Property("x")))
        self.build.setProperty("x", {"param_1": "param_1", "param_2": 2}, "here")
        self.expect_log_file(
            'log', f"URL: {url}?param_1=param_1&param_2=2\nStatus: 200\n ------ Content ------\nOK"
        )
        self.expect_log_file('content', "OK")
        self.expect_outcome(result=SUCCESS, state_string="Status code: 200")
        return self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.db.compression.protocol import CompressObjInterface
from buildbot.db.compression.protocol import CompressorInterface

try:
    import brotli

    MODE_TEXT = brotli.MODE_TEXT

    HAS_BROTLI = True
except ImportError:
    HAS_BROTLI = False
    MODE_TEXT = None


class BrotliCompressor(CompressorInterface):
    name = "br"
    available = HAS_BROTLI

    COMPRESS_QUALITY = 11
    MODE = MODE_TEXT

    @staticmethod
    def dumps(data: bytes) -> bytes:
        return brotli.compress(
            data,
            mode=BrotliCompressor.MODE,
            quality=BrotliCompressor.COMPRESS_QUALITY,
        )

    @staticmethod
    def read(data: bytes) -> bytes:
        return brotli.decompress(data)

    class CompressObj(CompressObjInterface):
        def __init__(self) -> None:
            self._create_compressobj()

        def _create_compressobj(self) -> None:
            self._compressobj = brotli.Compressor(
                mode=BrotliCompressor.MODE,
                quality=BrotliCompressor.COMPRESS_QUALITY,
            )

        def compress(self, data: bytes) -> bytes:
            return self._compressobj.process(data)

        def flush(self) -> bytes:
            try:
                return self._compressobj.finish()
            finally:
                # recreate compressobj so this instance can be re-used
                self._create_compressobj()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import cast

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable

    from twisted.internet.interfaces import IDelayedCall
    from twisted.internet.interfaces import IReactorTime

    from buildbot_worker.util.lineboundaries import LineInfo


class BufferManager:
    def __init__(
        self,
        reactor: IReactorTime,
        message_consumer: Callable[[list[tuple[str, Any]]], None],
        buffer_size: int,
        buffer_timeout: float,
    ) -> None:
        self._reactor = reactor
        self._buflen = 0
        self._buffered: list[tuple[str, Any]] = []
        self._buffer_size = buffer_size
        self._buffer_timeout = buffer_timeout
        self._send_message_timer: IDelayedCall | None = None
        self._message_consumer = message_consumer

    # TODO: Move to lineboundaries?
    def join_line_info(
        self,
        previous_line_info: LineInfo,
        new_line_info: LineInfo,
    ) -> LineInfo:
        previous_line_text = previous_line_info[0]
        len_previous_line_text = len(previous_line_text)
        new_line_text = previous_line_text + new_line_info[0]

        new_line_indexes = previous_line_info[1]
        for index in new_line_info[1]:
            new_line_indexes.append(len_previous_line_text + index)

        new_time_indexes = previous_line_info[2]
        for time in new_line_info[2]:
            new_time_indexes.append(time)

        return (new_line_text, new_line_indexes, new_time_indexes)

    def buffered_append_maybe_join_lines(
        self,
        logname: str,
        msg_data: tuple[str, LineInfo] | LineInfo,
    ) -> None:
        # if logname is the same as before: join message's line information with previous one
        if len(self._buffered) > 0 and self._buffered[-1][0] == logname:
            # different data format, when logname is 'log'
            # compare a log of first element of the data
            # second element is information about line
            # e.g. data = ('test_log', ('hello\n', [5], [0.0]))
            udpate_output = self._buffered[-1][1]
            if logname == "log":
                msg_data = cast("tuple[str, LineInfo]", msg_data)
                if udpate_output[0] == msg_data[0]:
                    joined_line_info = self.join_line_info(udpate_output[1], msg_data[1])
                    self._buffered[-1] = (logname, (msg_data[0], joined_line_info))
                    return
            else:
                msg_data = cast("LineInfo", msg_data)
                joined_line_info = self.join_line_info(udpate_output, msg_data)
                self._buffered[-1] = (logname, joined_line_info)
                return
        self._buffered.append((logname, msg_data))

    def setup_timeout(self) -> None:
        if not self._send_message_timer:
            self._send_message_timer = self._reactor.callLater(
                self._buffer_timeout, self.send_message_from_buffer
            )

    def append(self, logname: str, data: LineInfo | tuple[str, LineInfo] | int) -> None:
        # add data to the buffer for logname
        # keep appending to self._buffered until it gets longer than BUFFER_SIZE
        # which requires emptying the buffer by sending the message to the master

        is_log_message = logname in ("log", "stdout", "stderr", "header")

        if not is_log_message:
            len_data = 20
        else:
            if logname == "log":
                data = cast("tuple[str, LineInfo]", data)
                # different data format, when logname is 'log'
                # e.g. data = ('test_log', ('hello\n', [5], [0.0]))
                len_data = len(data[1][0]) + 8 * (len(data[1][1]) + len(data[1][2]))
            else:
                data = cast("LineInfo", data)
                len_data = len(data[0]) + 8 * (len(data[1]) + len(data[2]))

        space_left = self._buffer_size - self._buflen

        if len_data <= space_left:
            # fills the buffer with short lines
            if not is_log_message:
                self._buffered.append((logname, data))
            else:
                assert not isinstance(data, int)
                self.buffered_append_maybe_join_lines(logname, data)
            self._buflen += len_data
            self.setup_timeout()
            return

        self.send_message_from_buffer()

        if len_data <= self._buffer_size:
            self._buffered.append((logname, data))
            self._buflen += len_data
            self.setup_timeout()
            return

        if not is_log_message:
            self.send_message([(logname, data)])
            return

        if logname == "log":
            data = cast("tuple[str, LineInfo]", data)
            log = data[0]
            data = data[1]

        data = cast("LineInfo", data)

        pos_start = 0
        while pos_start < len(data[1]):
            # pos_end: which line is the last to be sent as a message (non-inclusive range)
            pos_end = pos_start + 1

            # Finds the range of lines to be sent:
            # pos_start - inclusive index of first line to be sent
            # pos_end - exclusive index of last line to be sent
            while pos_end <= len(data[1]):
                if pos_start == 0:
                    string_part_size = data[1][pos_end - 1] + 1
                else:
                    string_part_size = data[1][pos_end - 1] - (data[1][pos_start - 1])
                index_list_part_size = (pos_end - pos_start) * 8
                times_list_part_size = (pos_end - pos_start) * 8
                line_size = string_part_size + index_list_part_size + times_list_part_size

                if line_size <= self._buffer_size:
                    pos_end += 1
                else:
                    if pos_end > pos_start + 1:
                        pos_end -= 1
                    break

            if pos_end > len(data[1]):
                # no more lines are left to grab
                pos_end -= 1

            pos_substring_end = data[1][pos_end - 1] + 1
            if pos_start != 0:
                pos_substring_start = data[1][pos_start - 1] + 1
                line_info = (
                    data[0][pos_substring_start:pos_substring_end],
                    [index - pos_substring_start for index in data[1][pos_start:pos_end]],
                    data[2][pos_start:pos_end],
                )
            else:
                line_info = (data[0][:pos_substring_end], data[1][:pos_end], data[2][:pos_end])

            msg_data: tuple[str, LineInfo] | LineInfo
            if logname == "log":
                msg_data = (log, line_info)
            else:
                msg_data = line_info

            self.send_message([(logname, msg_data)])
            pos_start = pos_end

    def send_message_from_buffer(self) -> None:
        self.send_message(self._buffered)
        self._buffered = []
        self._buflen = 0

        if self._send_message_timer:
            if self._send_message_timer.active():
                self._send_message_timer.cancel()
            self._send_message_timer = None

    def send_message(
        self,
        data_to_send: list[tuple[str, tuple[str, LineInfo] | LineInfo | int]],
    ) -> None:
        if len(data_to_send) == 0:
            return
        self._message_consumer(data_to_send)

    def flush(self) -> None:
        if len(self._buffered) > 0:
            self.send_message_from_buffer()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from buildbot_worker.scripts.base import Options

workerTACTemplate = [
    """
import os

from buildbot_worker.bot import Worker
from twisted.application import service

basedir = %(basedir)r
rotateLength = %(log-size)d
maxRotatedFiles = %(log-count)s

# if this is a relocatable tac file, get the directory containing the TAC
if basedir == '.':
    import os.path
    basedir = os.path.abspath(os.path.dirname(__file__))

# note: this line is matched against to check that this is a worker
# directory; do not edit it.
application = service.Application('buildbot-worker')
""",
    """
from twisted.python.logfile import LogFile
from twisted.python.log import ILogObserver, FileLogObserver
logfile = LogFile.fromFullPath(
    os.path.join(basedir, "twistd.log"), rotateLength=rotateLength,
    maxRotatedFiles=maxRotatedFiles)
application.setComponent(ILogObserver, FileLogObserver(logfile).emit)
""",
    """
buildmaster_host = %(host)r
port = %(port)d
connection_string = None
""",
    """
buildmaster_host = None  # %(host)r
port = None  # %(port)d
connection_string = %(connection-string)r
""",
    """
workername = %(name)r
passwd = %(passwd)r
keepalive = %(keepalive)d
umask = %(umask)s
maxdelay = %(maxdelay)d
numcpus = %(numcpus)s
allow_shutdown = %(allow-shutdown)r
maxretries = %(maxretries)s
use_tls = %(use-tls)s
delete_leftover_dirs = %(delete-leftover-dirs)s
proxy_connection_string = %(proxy-connection-string)r
protocol = %(protocol)r

s = Worker(buildmaster_host, port, workername, passwd, basedir,
           keepalive, umask=umask, maxdelay=maxdelay,
           numcpus=numcpus, allow_shutdown=allow_shutdown,
           maxRetries=maxretries, protocol=protocol, useTls=use_tls,
           delete_leftover_dirs=delete_leftover_dirs,
           connection_string=connection_string,
           proxy_connection_string=proxy_connection_string)
s.setServiceParent(application)
""",
]


class CreateWorkerError(Exception):
    """
    Raised on errors while setting up worker directory.
    """


def _make_tac(config: Options) -> str:
    if config['relocatable']:
        config['basedir'] = '.'

    workerTAC = [workerTACTemplate[0]]

    if not config['no-logrotate']:
        workerTAC.append(workerTACTemplate[1])

    if not config['connection-string']:
        workerTAC.append(workerTACTemplate[2])
    else:
        workerTAC.append(workerTACTemplate[3])

    workerTAC.extend(workerTACTemplate[4:])

    return "".join(workerTAC) % config


def _makeBaseDir(basedir: str, quiet: bool) -> None:
    """
    Make worker base directory if needed.

    @param basedir: worker base directory relative path
    @param   quiet: if True, don't print info messages

    @raise CreateWorkerError: on error making base directory
    """
    if os.path.exists(basedir):
        if not quiet:
            print("updating existing installation")
        return

    if not quiet:
        print("mkdir", basedir)

    try:
        os.mkdir(basedir)
    except OSError as exception:
        raise CreateWorkerError(f"error creating directory {basedir}") from exception


def _makeBuildbotTac(basedir: str, tac_file_contents: str, quiet: bool) -> None:
    """
    Create buildbot.tac file. If buildbot.tac file already exists with
    different contents, create buildbot.tac.new instead.

    @param basedir: worker base directory relative path
    @param tac_file_contents: contents of buildbot.tac file to write
    @param quiet: if True, don't print info messages

    @raise CreateWorkerError: on error reading or writing tac file
    """
    tacfile = os.path.join(basedir, "buildbot.tac")

    if os.path.exists(tacfile):
        try:
            with open(tacfile) as f:
                oldcontents = f.read()
        except OSError as exception:
            raise CreateWorkerError(f"error reading {tacfile}") from exception

        if oldcontents == tac_file_contents:
            if not quiet:
                print("buildbot.tac already exists and is correct")
            return

        if not quiet:
            print("not touching existing buildbot.tac")
            print("creating buildbot.tac.new instead")

        tacfile = os.path.join(basedir, "buildbot.tac.new")

    try:
        with open(tacfile, "w") as f:
            f.write(tac_file_contents)
        os.chmod(tacfile, 0o600)
    except OSError as exception:
        raise CreateWorkerError(f"could not write {tacfile}") from exception


def _makeInfoFiles(basedir: str, quiet: bool) -> None:
    """
    Create info/* files inside basedir.

    @param basedir: worker base directory relative path
    @param   quiet: if True, don't print info messages

    @raise CreateWorkerError: on error making info directory or
                             writing info files
    """

    def createFile(path: str, file: str, contents: str) -> bool:
        filepath = os.path.join(path, file)

        if os.path.exists(filepath):
            return False

        if not quiet:
            print(
                "Creating {}, you need to edit it appropriately.".format(os.path.join("info", file))
            )

        try:
            open(filepath, "w").write(contents)
        except OSError as exception:
            raise CreateWorkerError(f"could not write {filepath}") from exception
        return True

    path = os.path.join(basedir, "info")
    if not os.path.exists(path):
        if not quiet:
            print("mkdir", path)
        try:
            os.mkdir(path)
        except OSError as exception:
            raise CreateWorkerError(f"error creating directory {path}") from exception

    # create 'info/admin' file
    created = createFile(path, "admin", "Your Name Here <admin@youraddress.invalid>\n")

    # create 'info/host' file
    created = createFile(path, "host", "Please put a description of this build host here\n")

    access_uri = os.path.join(path, "access_uri")

    if not os.path.exists(access_uri):
        if not quiet:
            print("Not creating {} - add it if you wish".format(os.path.join("info", "access_uri")))

    if created and not quiet:
        print(f"Please edit the files in {path} appropriately.")


def createWorker(config: Options) -> int:
    basedir = config['basedir']
    quiet = config['quiet']

    contents = _make_tac(config)

    try:
        _makeBaseDir(basedir, quiet)
        _makeBuildbotTac(basedir, contents, quiet)
        _makeInfoFiles(basedir, quiet)
    except CreateWorkerError as exception:
        print("{}\nfailed to configure worker in {}".format(exception, config['basedir']))
        return 1

    if not quiet:
        print(f"worker configured in {basedir}")

    return 0



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import test_result_sets
from buildbot.db.test_result_sets import TestResultSetModel
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util import interfaces
from buildbot.util.twisted import async_to_deferred


class TestResultSetEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = test_result_sets.TestResultSetEndpoint
    resourceTypeClass = test_result_sets.TestResultSet

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        yield self.master.db.insert_test_data([
            fakedb.Worker(id=47, name='linux'),
            fakedb.Buildset(id=20),
            fakedb.Builder(id=88, name='b1'),
            fakedb.BuildRequest(id=41, buildsetid=20, builderid=88),
            fakedb.Master(id=88),
            fakedb.Build(
                id=30, buildrequestid=41, number=7, masterid=88, builderid=88, workerid=47
            ),
            fakedb.Step(id=131, number=132, name='step132', buildid=30),
            fakedb.TestResultSet(
                id=13,
                builderid=88,
                buildid=30,
                stepid=131,
                description='desc',
                category='cat',
                value_unit='ms',
                complete=1,
            ),
        ])

    @defer.inlineCallbacks
    def test_get_existing_result_set(self):
        result = yield self.callGet(('test_result_sets', 13))
        self.validateData(result)
        self.assertEqual(
            result,
            {
                'test_result_setid': 13,
                'builderid': 88,
                'buildid': 30,
                'stepid': 131,
                'description': 'desc',
                'category': 'cat',
                'value_unit': 'ms',
                'tests_passed': None,
                'tests_failed': None,
                'complete': True,
            },
        )

    @defer.inlineCallbacks
    def test_get_missing_result_set(self):
        results = yield self.callGet(('test_result_sets', 14))
        self.assertIsNone(results)


class TestResultSetsEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = test_result_sets.TestResultSetsEndpoint
    resourceTypeClass = test_result_sets.TestResultSet

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        yield self.master.db.insert_test_data([
            fakedb.Worker(id=47, name='linux'),
            fakedb.Buildset(id=20),
            fakedb.Builder(id=88, name='b1'),
            fakedb.BuildRequest(id=41, buildsetid=20, builderid=88),
            fakedb.Master(id=88),
            fakedb.Build(
                id=30, buildrequestid=41, number=7, masterid=88, builderid=88, workerid=47
            ),
            fakedb.Step(id=131, number=132, name='step132', buildid=30),
            fakedb.TestResultSet(
                id=13,
                builderid=88,
                buildid=30,
                stepid=131,
                description='desc',
                category='cat',
                value_unit='ms',
                complete=1,
            ),
            fakedb.TestResultSet(
                id=14,
                builderid=88,
                buildid=30,
                stepid=131,
                description='desc',
                category='cat',
                value_unit='ms',
                complete=1,
            ),
        ])

    @defer.inlineCallbacks
    def test_get_result_sets_all(self):
        results = yield self.callGet(('test_result_sets',))
        for result in results:
            self.validateData(result)
        self.assertEqual([r['test_result_setid'] for r in results], [13, 14])

    @defer.inlineCallbacks
    def test_get_result_sets_builders_builderid(self):
        results = yield self.callGet(('builders', 88, 'test_result_sets'))
        for result in results:
            self.validateData(result)
        self.assertEqual([r['test_result_setid'] for r in results], [13, 14])

    @defer.inlineCallbacks
    def test_get_result_sets_builders_buildername(self):
        results = yield self.callGet(('builders', 'b1', 'test_result_sets'))
        for result in results:
            self.validateData(result)
        self.assertEqual([r['test_result_setid'] for r in results], [13, 14])

    @defer.inlineCallbacks
    def test_get_result_sets_builds_buildid(self):
        results = yield self.callGet(('builds', 30, 'test_result_sets'))
        for result in results:
            self.validateData(result)
        self.assertEqual([r['test_result_setid'] for r in results], [13, 14])

    @defer.inlineCallbacks
    def test_get_result_sets_steps_stepid(self):
        results = yield self.callGet(('steps', 131, 'test_result_sets'))
        for result in results:
            self.validateData(result)
        self.assertEqual([r['test_result_setid'] for r in results], [13, 14])


class TestResultSetsFromCommitRangeEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = test_result_sets.TestResultSetsFromCommitRangeEndpoint
    resourceTypeClass = test_result_sets.TestResultSet

    @async_to_deferred
    async def setUp(self) -> None:  # type: ignore[override]
        await self.setUpEndpoint()

        master_id = fakedb.FakeDBConnector.MASTER_ID
        await self.master.db.insert_test_data([
            fakedb.Master(id=master_id),
            fakedb.Worker(id=47, name='linux'),
            fakedb.Project(id=100),
            fakedb.Codebase(id=200, projectid=100),
            fakedb.CodebaseCommit(id=300, codebaseid=200, revision='rev300'),
            fakedb.CodebaseCommit(id=301, codebaseid=200, revision='rev301', parent_commitid=300),
            fakedb.CodebaseCommit(id=302, codebaseid=200, revision='rev302', parent_commitid=301),
            fakedb.CodebaseCommit(id=303, codebaseid=200, revision='rev303', parent_commitid=302),
            fakedb.CodebaseCommit(id=304, codebaseid=200, revision='rev304', parent_commitid=303),
            fakedb.Buildset(id=4000),
            fakedb.Buildset(id=4010),
            fakedb.Buildset(id=4020),
            fakedb.Buildset(id=4021),
            fakedb.Buildset(id=4040),
            fakedb.SourceStamp(id=5000, revision='rev300'),
            fakedb.SourceStamp(id=5010, revision='rev301'),
            fakedb.SourceStamp(id=5020, revision='rev302'),
            fakedb.SourceStamp(id=5040, revision='rev304'),
            fakedb.BuildsetSourceStamp(id=6000, buildsetid=4000, sourcestampid=5000),
            fakedb.BuildsetSourceStamp(id=6001, buildsetid=4010, sourcestampid=5010),
            fakedb.BuildsetSourceStamp(id=6002, buildsetid=4020, sourcestampid=5020),
            fakedb.BuildsetSourceStamp(id=6003, buildsetid=4021, sourcestampid=5020),
            fakedb.BuildsetSourceStamp(id=6004, buildsetid=4040, sourcestampid=5040),
            fakedb.Builder(id=400, name='b1'),
            fakedb.BuildRequest(id=7000, buildsetid=4000, builderid=400),
            fakedb.BuildRequest(id=7010, buildsetid=4010, builderid=400),
            fakedb.BuildRequest(id=7020, buildsetid=4020, builderid=400),
            fakedb.BuildRequest(id=7021, buildsetid=4021, builderid=400),
            fakedb.BuildRequest(id=7040, buildsetid=4040, builderid=400),
            fakedb.Build(
                id=8000, buildrequestid=7000, masterid=master_id, builderid=400, workerid=47
            ),
            fakedb.Build(
                id=8010, buildrequestid=7010, masterid=master_id, builderid=400, workerid=47
            ),
            fakedb.Build(
                id=8020, buildrequestid=7020, masterid=master_id, builderid=400, workerid=47
            ),
            fakedb.Build(
                id=8021, buildrequestid=7021, masterid=master_id, builderid=400, workerid=47
            ),
            fakedb.Build(
                id=8040, buildrequestid=7040, masterid=master_id, builderid=400, workerid=47
            ),
            fakedb.Step(id=9000, buildid=8000),
            fakedb.Step(id=9010, buildid=8010),
            fakedb.Step(id=9020, buildid=8020),
            fakedb.Step(id=9021, buildid=8021),
            fakedb.Step(id=9040, buildid=8040),
            fakedb.TestResultSet(id=10000, builderid=400, buildid=8000, stepid=9000),
            fakedb.TestResultSet(id=10010, builderid=400, buildid=8010, stepid=9010),
            fakedb.TestResultSet(id=10020, builderid=400, buildid=8020, stepid=9020),
            fakedb.TestResultSet(id=10021, builderid=400, buildid=8021, stepid=9021),
            fakedb.TestResultSet(id=10040, builderid=400, buildid=8040, stepid=9040),
        ])

    @async_to_deferred
    async def test_get_full_range(self) -> None:
        results = await self.callGet((
            'codebases',
            200,
            'commit_range',
            300,
            304,
            'test_result_sets',
        ))
        for result in results:
            self.validateData(result)
        self.assertEqual(
            [r['test_result_setid'] for r in results], [10000, 10010, 10020, 10021, 10040]
        )

    @async_to_deferred
    async def test_get_opposite_range(self) -> None:
        results = await self.callGet((
            'codebases',
            200,
            'commit_range',
            304,
            300,
            'test_result_sets',
        ))
        self.assertEqual(results, [])


class TestResultSet(TestReactorMixin, interfaces.InterfaceTests, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantMq=True, wantDb=True, wantData=True)
        yield self.master.db.insert_test_data([
            fakedb.Master(id=1),
            fakedb.Worker(id=1, name='example-worker'),
            fakedb.Builder(id=1),
            fakedb.Buildset(id=1),
            fakedb.BuildRequest(
                id=1,
                buildsetid=1,
                builderid=1,
            ),
            fakedb.Build(
                id=2,
                number=1,
                buildrequestid=1,
                builderid=1,
                workerid=1,
                masterid=1,
            ),
            fakedb.Step(
                id=3,
                number=1,
                name='step1',
                buildid=2,
            ),
        ])
        self.rtype = test_result_sets.TestResultSet(self.master)

    def test_signature_add_test_result_set(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.addTestResultSet, self.rtype.addTestResultSet
        )
        def addTestResultSet(self, builderid, buildid, stepid, description, category, value_unit):
            pass

    def test_signature_complete_test_result_set(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.completeTestResultSet, self.rtype.completeTestResultSet
        )
        def completeTestResultSet(self, test_result_setid, tests_passed=None, tests_failed=None):
            pass

    @defer.inlineCallbacks
    def test_add_test_result_set(self):
        test_result_setid = yield self.rtype.addTestResultSet(
            builderid=1, buildid=2, stepid=3, description='desc', category='cat4', value_unit='ms'
        )

        msg_body = {
            'test_result_setid': test_result_setid,
            'builderid': 1,
            'buildid': 2,
            'stepid': 3,
            'description': 'desc',
            'category': 'cat4',
            'value_unit': 'ms',
            'tests_passed': None,
            'tests_failed': None,
            'complete': False,
        }

        self.master.mq.assertProductions([
            (('test_result_sets', str(test_result_setid), 'new'), msg_body),
        ])

        result = yield self.master.db.test_result_sets.getTestResultSet(test_result_setid)
        self.assertEqual(
            result,
            TestResultSetModel(
                id=test_result_setid,
                builderid=1,
                buildid=2,
                stepid=3,
                description='desc',
                category='cat4',
                value_unit='ms',
                tests_passed=None,
                tests_failed=None,
                complete=False,
            ),
        )

    @defer.inlineCallbacks
    def test_complete_test_result_set_no_results(self):
        test_result_setid = yield self.master.db.test_result_sets.addTestResultSet(
            builderid=1, buildid=2, stepid=3, description='desc', category='cat4', value_unit='ms'
        )

        yield self.rtype.completeTestResultSet(test_result_setid)

        msg_body = {
            'test_result_setid': test_result_setid,
            'builderid': 1,
            'buildid': 2,
            'stepid': 3,
            'description': 'desc',
            'category': 'cat4',
            'value_unit': 'ms',
            'tests_passed': None,
            'tests_failed': None,
            'complete': True,
        }

        self.master.mq.assertProductions([
            (('test_result_sets', str(test_result_setid), 'completed'), msg_body),
        ])

        result = yield self.master.db.test_result_sets.getTestResultSet(test_result_setid)
        self.assertEqual(
            result,
            TestResultSetModel(
                id=test_result_setid,
                builderid=1,
                buildid=2,
                stepid=3,
                description='desc',
                category='cat4',
                value_unit='ms',
                tests_passed=None,
                tests_failed=None,
                complete=True,
            ),
        )

    @defer.inlineCallbacks
    def test_complete_test_result_set_with_results(self):
        test_result_setid = yield self.master.db.test_result_sets.addTestResultSet(
            builderid=1, buildid=2, stepid=3, description='desc', category='cat4', value_unit='ms'
        )

        yield self.rtype.completeTestResultSet(test_result_setid, tests_passed=12, tests_failed=34)

        msg_body = {
            'test_result_setid': test_result_setid,
            'builderid': 1,
            'buildid': 2,
            'stepid': 3,
            'description': 'desc',
            'category': 'cat4',
            'value_unit': 'ms',
            'tests_passed': 12,
            'tests_failed': 34,
            'complete': True,
        }

        self.master.mq.assertProductions([
            (('test_result_sets', str(test_result_setid), 'completed'), msg_body),
        ])

        result = yield self.master.db.test_result_sets.getTestResultSet(test_result_setid)
        self.assertEqual(
            result,
            TestResultSetModel(
                id=test_result_setid,
                builderid=1,
                buildid=2,
                stepid=3,
                description='desc',
                category='cat4',
                value_unit='ms',
                tests_passed=12,
                tests_failed=34,
                complete=True,
            ),
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import workers
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import querylog


def workerKey(worker: workers.WorkerModel):
    return worker.id


def configuredOnKey(worker: workers.BuilderMasterModel):
    return (worker.builderid, worker.masterid)


class Tests(TestReactorMixin, unittest.TestCase, querylog.SqliteMaxVariableMixin):
    # common sample data

    baseRows = [
        fakedb.Master(id=10, name='m10'),
        fakedb.Master(id=11, name='m11'),
        fakedb.Builder(id=20, name='a'),
        fakedb.Builder(id=21, name='b'),
        fakedb.Builder(id=22, name='c'),
        fakedb.Worker(id=30, name='zero'),
        fakedb.Worker(id=31, name='one'),
    ]

    multipleMasters = [
        fakedb.BuilderMaster(id=12, builderid=20, masterid=10),
        fakedb.BuilderMaster(id=13, builderid=21, masterid=10),
        fakedb.BuilderMaster(id=14, builderid=20, masterid=11),
        fakedb.BuilderMaster(id=15, builderid=22, masterid=11),
        fakedb.BuilderMaster(id=16, builderid=22, masterid=10),
        fakedb.ConfiguredWorker(id=3012, workerid=30, buildermasterid=12),
        fakedb.ConfiguredWorker(id=3013, workerid=30, buildermasterid=13),
        fakedb.ConfiguredWorker(id=3014, workerid=30, buildermasterid=14),
        fakedb.ConfiguredWorker(id=3114, workerid=31, buildermasterid=14),
        fakedb.ConfiguredWorker(id=3115, workerid=31, buildermasterid=15),
        fakedb.ConnectedWorker(id=3010, workerid=30, masterid=10),
        fakedb.ConnectedWorker(id=3111, workerid=31, masterid=11),
    ]

    # sample worker data, with id's avoiding the postgres id sequence

    BOGUS_NAME = 'bogus'

    W1_NAME = "w1"
    W1_ID = 100
    W1_INFO = {'a': 1}
    worker1_rows = [
        fakedb.Worker(id=W1_ID, name=W1_NAME, info=W1_INFO),
    ]

    W2_NAME = "w2"
    W2_ID = 200
    W2_INFO = {'a': 1, 'b': 2}
    worker2_rows = [
        fakedb.Worker(id=W2_ID, name=W2_NAME, info=W2_INFO),
    ]

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def test_findWorkerId_insert(self):
        id = yield self.db.workers.findWorkerId(name="xyz")
        worker = yield self.db.workers.getWorker(workerid=id)
        self.assertEqual(worker.name, 'xyz')
        self.assertEqual(worker.workerinfo, {})

    @defer.inlineCallbacks
    def test_findWorkerId_existing(self):
        yield self.db.insert_test_data(self.baseRows)
        id = yield self.db.workers.findWorkerId(name="one")
        self.assertEqual(id, 31)

    @defer.inlineCallbacks
    def test_getWorker_no_such(self):
        yield self.db.insert_test_data(self.baseRows)
        workerdict = yield self.db.workers.getWorker(workerid=99)
        self.assertEqual(workerdict, None)

    @defer.inlineCallbacks
    def test_getWorker_by_name_no_such(self):
        yield self.db.insert_test_data(self.baseRows)
        workerdict = yield self.db.workers.getWorker(name='NOSUCH')
        self.assertEqual(workerdict, None)

    @defer.inlineCallbacks
    def test_getWorker_not_configured(self):
        yield self.db.insert_test_data(self.baseRows)
        workerdict = yield self.db.workers.getWorker(workerid=30)
        self.assertIsInstance(workerdict, workers.WorkerModel)
        self.assertEqual(
            workerdict,
            workers.WorkerModel(
                id=30,
                name='zero',
                workerinfo={'a': 'b'},
                paused=False,
                pause_reason=None,
                graceful=False,
                connected_to=[],
                configured_on=[],
            ),
        )

    @defer.inlineCallbacks
    def test_getWorker_connected_not_configured(self):
        yield self.db.insert_test_data([
            *self.baseRows,
            # the worker is connected to this master, but not configured.
            # weird, but the DB should represent it.
            fakedb.Worker(id=32, name='two'),
            fakedb.ConnectedWorker(workerid=32, masterid=11),
        ])
        workerdict = yield self.db.workers.getWorker(workerid=32)
        self.assertIsInstance(workerdict, workers.WorkerModel)
        self.assertEqual(
            workerdict,
            workers.WorkerModel(
                id=32,
                name='two',
                workerinfo={'a': 'b'},
                paused=False,
                pause_reason=None,
                graceful=False,
                connected_to=[11],
                configured_on=[],
            ),
        )

    @defer.inlineCallbacks
    def test_getWorker_multiple_connections(self):
        yield self.db.insert_test_data([
            *self.baseRows,
            # the worker is connected to two masters at once.
            # weird, but the DB should represent it.
            fakedb.Worker(id=32, name='two'),
            fakedb.ConnectedWorker(workerid=32, masterid=10),
            fakedb.ConnectedWorker(workerid=32, masterid=11),
            fakedb.BuilderMaster(id=24, builderid=20, masterid=10),
            fakedb.BuilderMaster(id=25, builderid=20, masterid=11),
            fakedb.ConfiguredWorker(workerid=32, buildermasterid=24),
            fakedb.ConfiguredWorker(workerid=32, buildermasterid=25),
        ])
        workerdict = yield self.db.workers.getWorker(workerid=32)
        self.assertIsInstance(workerdict, workers.WorkerModel)
        self.assertEqual(
            workerdict,
            workers.WorkerModel(
                id=32,
                name='two',
                workerinfo={'a': 'b'},
                paused=False,
                pause_reason=None,
                graceful=False,
                connected_to=[10, 11],
                configured_on=[
                    workers.BuilderMasterModel(builderid=20, masterid=10),
                    workers.BuilderMasterModel(builderid=20, masterid=11),
                ],
            ),
        )

    @defer.inlineCallbacks
    def test_getWorker_by_name_not_configured(self):
        yield self.db.insert_test_data(self.baseRows)
        workerdict = yield self.db.workers.getWorker(name='zero')
        self.assertIsInstance(workerdict, workers.WorkerModel)
        self.assertEqual(
            workerdict,
            workers.WorkerModel(
                id=30,
                name='zero',
                workerinfo={'a': 'b'},
                paused=False,
                pause_reason=None,
                graceful=False,
                connected_to=[],
                configured_on=[],
            ),
        )

    @defer.inlineCallbacks
    def test_getWorker_not_connected(self):
        yield self.db.insert_test_data([
            *self.baseRows,
            fakedb.BuilderMaster(id=12, builderid=20, masterid=10),
            fakedb.ConfiguredWorker(workerid=30, buildermasterid=12),
        ])
        workerdict = yield self.db.workers.getWorker(workerid=30)
        self.assertIsInstance(workerdict, workers.WorkerModel)
        self.assertEqual(
            workerdict,
            workers.WorkerModel(
                id=30,
                name='zero',
                workerinfo={'a': 'b'},
                paused=False,
                pause_reason=None,
                graceful=False,
                configured_on=[workers.BuilderMasterModel(masterid=10, builderid=20)],
                connected_to=[],
            ),
        )

    @defer.inlineCallbacks
    def test_getWorker_connected(self):
        yield self.db.insert_test_data([
            *self.baseRows,
            fakedb.BuilderMaster(id=12, builderid=20, masterid=10),
            fakedb.ConfiguredWorker(workerid=30, buildermasterid=12),
            fakedb.ConnectedWorker(workerid=30, masterid=10),
        ])
        workerdict = yield self.db.workers.getWorker(workerid=30)
        self.assertIsInstance(workerdict, workers.WorkerModel)
        self.assertEqual(
            workerdict,
            workers.WorkerModel(
                id=30,
                name='zero',
                workerinfo={'a': 'b'},
                paused=False,
                pause_reason=None,
                graceful=False,
                configured_on=[workers.BuilderMasterModel(masterid=10, builderid=20)],
                connected_to=[10],
            ),
        )

    @defer.inlineCallbacks
    def test_getWorker_with_multiple_masters(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        workerdict = yield self.db.workers.getWorker(workerid=30)
        self.assertIsInstance(workerdict, workers.WorkerModel)
        workerdict.configured_on = sorted(workerdict.configured_on, key=configuredOnKey)
        self.assertEqual(
            workerdict,
            workers.WorkerModel(
                id=30,
                name='zero',
                workerinfo={'a': 'b'},
                paused=False,
                pause_reason=None,
                graceful=False,
                configured_on=sorted(
                    [
                        workers.BuilderMasterModel(masterid=10, builderid=20),
                        workers.BuilderMasterModel(masterid=10, builderid=21),
                        workers.BuilderMasterModel(masterid=11, builderid=20),
                    ],
                    key=configuredOnKey,
                ),
                connected_to=[10],
            ),
        )

    @defer.inlineCallbacks
    def test_getWorker_with_multiple_masters_builderid(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        workerdict = yield self.db.workers.getWorker(workerid=30, builderid=20)
        self.assertIsInstance(workerdict, workers.WorkerModel)
        workerdict.configured_on = sorted(workerdict.configured_on, key=configuredOnKey)
        self.assertEqual(
            workerdict,
            workers.WorkerModel(
                id=30,
                name='zero',
                workerinfo={'a': 'b'},
                paused=False,
                pause_reason=None,
                graceful=False,
                configured_on=sorted(
                    [
                        workers.BuilderMasterModel(builderid=20, masterid=10),
                        workers.BuilderMasterModel(builderid=20, masterid=11),
                    ],
                    key=configuredOnKey,
                ),
                connected_to=[10],
            ),
        )

    @defer.inlineCallbacks
    def test_getWorker_with_multiple_masters_masterid(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        workerdict = yield self.db.workers.getWorker(workerid=30, masterid=11)
        self.assertIsInstance(workerdict, workers.WorkerModel)
        self.assertEqual(
            workerdict,
            workers.WorkerModel(
                id=30,
                name='zero',
                workerinfo={'a': 'b'},
                paused=False,
                pause_reason=None,
                graceful=False,
                configured_on=[
                    workers.BuilderMasterModel(builderid=20, masterid=11),
                ],
                connected_to=[],
            ),
        )

    @defer.inlineCallbacks
    def test_getWorker_with_multiple_masters_builderid_masterid(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        workerdict = yield self.db.workers.getWorker(workerid=30, builderid=20, masterid=11)
        self.assertIsInstance(workerdict, workers.WorkerModel)
        self.assertEqual(
            workerdict,
            workers.WorkerModel(
                id=30,
                name='zero',
                workerinfo={'a': 'b'},
                paused=False,
                pause_reason=None,
                graceful=False,
                configured_on=[
                    workers.BuilderMasterModel(builderid=20, masterid=11),
                ],
                connected_to=[],
            ),
        )

    @defer.inlineCallbacks
    def test_getWorker_by_name_with_multiple_masters_builderid_masterid(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        workerdict = yield self.db.workers.getWorker(name='zero', builderid=20, masterid=11)
        self.assertIsInstance(workerdict, workers.WorkerModel)
        self.assertEqual(
            workerdict,
            workers.WorkerModel(
                id=30,
                name='zero',
                workerinfo={'a': 'b'},
                paused=False,
                pause_reason=None,
                graceful=False,
                configured_on=[
                    workers.BuilderMasterModel(masterid=11, builderid=20),
                ],
                connected_to=[],
            ),
        )

    @defer.inlineCallbacks
    def test_getWorkers_no_config(self):
        yield self.db.insert_test_data(self.baseRows)
        workerdicts = yield self.db.workers.getWorkers()

        for workerdict in workerdicts:
            self.assertIsInstance(workerdict, workers.WorkerModel)

        self.assertEqual(
            sorted(workerdicts, key=workerKey),
            sorted(
                [
                    workers.WorkerModel(
                        id=30,
                        name='zero',
                        workerinfo={'a': 'b'},
                        paused=False,
                        pause_reason=None,
                        graceful=False,
                        configured_on=[],
                        connected_to=[],
                    ),
                    workers.WorkerModel(
                        id=31,
                        name='one',
                        workerinfo={'a': 'b'},
                        paused=False,
                        pause_reason=None,
                        graceful=False,
                        configured_on=[],
                        connected_to=[],
                    ),
                ],
                key=workerKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getWorkers_with_config(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        workerdicts = yield self.db.workers.getWorkers()
        for workerdict in workerdicts:
            self.assertIsInstance(workerdict, workers.WorkerModel)
            workerdict.configured_on = sorted(workerdict.configured_on, key=configuredOnKey)
        self.assertEqual(
            sorted(workerdicts, key=workerKey),
            sorted(
                [
                    workers.WorkerModel(
                        id=30,
                        name='zero',
                        workerinfo={'a': 'b'},
                        paused=False,
                        pause_reason=None,
                        graceful=False,
                        configured_on=sorted(
                            [
                                workers.BuilderMasterModel(builderid=20, masterid=10),
                                workers.BuilderMasterModel(builderid=21, masterid=10),
                                workers.BuilderMasterModel(builderid=20, masterid=11),
                            ],
                            key=configuredOnKey,
                        ),
                        connected_to=[10],
                    ),
                    workers.WorkerModel(
                        id=31,
                        name='one',
                        workerinfo={'a': 'b'},
                        paused=False,
                        pause_reason=None,
                        graceful=False,
                        configured_on=sorted(
                            [
                                workers.BuilderMasterModel(builderid=20, masterid=11),
                                workers.BuilderMasterModel(builderid=22, masterid=11),
                            ],
                            key=configuredOnKey,
                        ),
                        connected_to=[11],
                    ),
                ],
                key=workerKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getWorkers_empty(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        workerdicts = yield self.db.workers.getWorkers(masterid=11, builderid=21)
        for workerdict in workerdicts:
            self.assertIsInstance(workerdict, workers.WorkerModel)
            workerdict.configured_on = sorted(workerdict.configured_on, key=configuredOnKey)
        self.assertEqual(sorted(workerdicts, key=workerKey), [])

    @defer.inlineCallbacks
    def test_getWorkers_with_config_builderid(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        workerdicts = yield self.db.workers.getWorkers(builderid=20)
        for workerdict in workerdicts:
            self.assertIsInstance(workerdict, workers.WorkerModel)
            workerdict.configured_on = sorted(workerdict.configured_on, key=configuredOnKey)
        self.assertEqual(
            sorted(workerdicts, key=workerKey),
            sorted(
                [
                    workers.WorkerModel(
                        id=30,
                        name='zero',
                        workerinfo={'a': 'b'},
                        paused=False,
                        pause_reason=None,
                        graceful=False,
                        configured_on=sorted(
                            [
                                workers.BuilderMasterModel(builderid=20, masterid=10),
                                workers.BuilderMasterModel(builderid=20, masterid=11),
                            ],
                            key=configuredOnKey,
                        ),
                        connected_to=[10],
                    ),
                    workers.WorkerModel(
                        id=31,
                        name='one',
                        workerinfo={'a': 'b'},
                        paused=False,
                        pause_reason=None,
                        graceful=False,
                        configured_on=sorted(
                            [
                                workers.BuilderMasterModel(builderid=20, masterid=11),
                            ],
                            key=configuredOnKey,
                        ),
                        connected_to=[11],
                    ),
                ],
                key=workerKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getWorkers_with_config_masterid_10(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        workerdicts = yield self.db.workers.getWorkers(masterid=10)
        for workerdict in workerdicts:
            self.assertIsInstance(workerdict, workers.WorkerModel)
            workerdict.configured_on = sorted(workerdict.configured_on, key=configuredOnKey)
        self.assertEqual(
            sorted(workerdicts, key=workerKey),
            sorted(
                [
                    workers.WorkerModel(
                        id=30,
                        name='zero',
                        workerinfo={'a': 'b'},
                        paused=False,
                        pause_reason=None,
                        graceful=False,
                        configured_on=sorted(
                            [
                                workers.BuilderMasterModel(builderid=20, masterid=10),
                                workers.BuilderMasterModel(builderid=21, masterid=10),
                            ],
                            key=configuredOnKey,
                        ),
                        connected_to=[10],
                    ),
                ],
                key=workerKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getWorkers_with_config_masterid_11(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        workerdicts = yield self.db.workers.getWorkers(masterid=11)
        for workerdict in workerdicts:
            self.assertIsInstance(workerdict, workers.WorkerModel)
            workerdict.configured_on = sorted(workerdict.configured_on, key=configuredOnKey)
        self.assertEqual(
            sorted(workerdicts, key=workerKey),
            sorted(
                [
                    workers.WorkerModel(
                        id=30,
                        name='zero',
                        workerinfo={'a': 'b'},
                        paused=False,
                        pause_reason=None,
                        graceful=False,
                        configured_on=sorted(
                            [
                                workers.BuilderMasterModel(builderid=20, masterid=11),
                            ],
                            key=configuredOnKey,
                        ),
                        connected_to=[],
                    ),
                    workers.WorkerModel(
                        id=31,
                        name='one',
                        workerinfo={'a': 'b'},
                        paused=False,
                        pause_reason=None,
                        graceful=False,
                        configured_on=sorted(
                            [
                                workers.BuilderMasterModel(builderid=20, masterid=11),
                                workers.BuilderMasterModel(builderid=22, masterid=11),
                            ],
                            key=configuredOnKey,
                        ),
                        connected_to=[11],
                    ),
                ],
                key=workerKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getWorkers_with_config_masterid_11_builderid_22(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        workerdicts = yield self.db.workers.getWorkers(masterid=11, builderid=22)
        for workerdict in workerdicts:
            self.assertIsInstance(workerdict, workers.WorkerModel)
            workerdict.configured_on = sorted(workerdict.configured_on, key=configuredOnKey)
        self.assertEqual(
            sorted(workerdicts, key=workerKey),
            sorted(
                [
                    workers.WorkerModel(
                        id=31,
                        name='one',
                        workerinfo={'a': 'b'},
                        paused=False,
                        pause_reason=None,
                        graceful=False,
                        configured_on=sorted(
                            [
                                workers.BuilderMasterModel(builderid=22, masterid=11),
                            ],
                            key=configuredOnKey,
                        ),
                        connected_to=[11],
                    ),
                ],
                key=workerKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getWorkers_with_paused(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        yield self.db.workers.set_worker_paused(31, paused=True, pause_reason="reason")
        yield self.db.workers.set_worker_graceful(31, graceful=False)
        workerdicts = yield self.db.workers.getWorkers(paused=True)
        for workerdict in workerdicts:
            self.assertIsInstance(workerdict, workers.WorkerModel)
            workerdict.configured_on = []
        self.assertEqual(
            workerdicts,
            [
                workers.WorkerModel(
                    id=31,
                    name='one',
                    workerinfo={'a': 'b'},
                    paused=True,
                    pause_reason="reason",
                    graceful=False,
                    configured_on=[],
                    connected_to=[11],
                ),
            ],
        )

    @defer.inlineCallbacks
    def test_getWorkers_with_graceful(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)
        yield self.db.workers.set_worker_paused(31, paused=False)
        yield self.db.workers.set_worker_graceful(31, graceful=True)
        workerdicts = yield self.db.workers.getWorkers(graceful=True)
        for workerdict in workerdicts:
            self.assertIsInstance(workerdict, workers.WorkerModel)
            workerdict.configured_on = []
        self.assertEqual(
            workerdicts,
            [
                workers.WorkerModel(
                    id=31,
                    name='one',
                    workerinfo={'a': 'b'},
                    paused=False,
                    pause_reason=None,
                    graceful=True,
                    configured_on=[],
                    connected_to=[11],
                ),
            ],
        )

    @defer.inlineCallbacks
    def test_workerConnected_existing(self):
        yield self.db.insert_test_data(self.baseRows + self.worker1_rows)

        NEW_INFO = {'other': [1, 2, 3]}

        yield self.db.workers.workerConnected(workerid=self.W1_ID, masterid=11, workerinfo=NEW_INFO)

        w = yield self.db.workers.getWorker(self.W1_ID)
        self.assertEqual(
            w,
            workers.WorkerModel(
                id=self.W1_ID,
                name=self.W1_NAME,
                workerinfo=NEW_INFO,
                paused=False,
                pause_reason=None,
                graceful=False,
                configured_on=[],
                connected_to=[11],
            ),
        )

    @defer.inlineCallbacks
    def test_workerConnected_already_connected(self):
        yield self.db.insert_test_data(
            self.baseRows
            + self.worker1_rows
            + [
                fakedb.ConnectedWorker(id=888, workerid=self.W1_ID, masterid=11),
            ]
        )
        yield self.db.workers.workerConnected(workerid=self.W1_ID, masterid=11, workerinfo={})

        w = yield self.db.workers.getWorker(self.W1_ID)
        self.assertEqual(w.connected_to, [11])

    @defer.inlineCallbacks
    def test_workerDisconnected(self):
        yield self.db.insert_test_data(
            self.baseRows
            + self.worker1_rows
            + [
                fakedb.ConnectedWorker(id=888, workerid=self.W1_ID, masterid=10),
                fakedb.ConnectedWorker(id=889, workerid=self.W1_ID, masterid=11),
            ]
        )
        yield self.db.workers.workerDisconnected(workerid=self.W1_ID, masterid=11)

        w = yield self.db.workers.getWorker(self.W1_ID)
        self.assertEqual(w.connected_to, [10])

    @defer.inlineCallbacks
    def test_workerDisconnected_already_disconnected(self):
        yield self.db.insert_test_data(self.baseRows + self.worker1_rows)
        yield self.db.workers.workerDisconnected(workerid=self.W1_ID, masterid=11)

        w = yield self.db.workers.getWorker(self.W1_ID)
        self.assertEqual(w.connected_to, [])

    @defer.inlineCallbacks
    def test_set_worker_paused_existing(self):
        yield self.db.insert_test_data(self.baseRows + self.worker1_rows)

        yield self.db.workers.set_worker_paused(self.W1_ID, False, None)

        w = yield self.db.workers.getWorker(self.W1_ID)
        self.assertEqual(
            w,
            workers.WorkerModel(
                id=self.W1_ID,
                name=self.W1_NAME,
                workerinfo=self.W1_INFO,
                paused=False,
                pause_reason=None,
                graceful=False,
                configured_on=[],
                connected_to=[],
            ),
        )

        yield self.db.workers.set_worker_paused(self.W1_ID, True, "reason")

        w = yield self.db.workers.getWorker(self.W1_ID)
        self.assertEqual(
            w,
            workers.WorkerModel(
                id=self.W1_ID,
                name=self.W1_NAME,
                workerinfo=self.W1_INFO,
                paused=True,
                pause_reason="reason",
                graceful=False,
                configured_on=[],
                connected_to=[],
            ),
        )

    @defer.inlineCallbacks
    def test_set_worker_graceful_existing(self):
        yield self.db.insert_test_data(self.baseRows + self.worker1_rows)

        yield self.db.workers.set_worker_graceful(self.W1_ID, False)

        w = yield self.db.workers.getWorker(self.W1_ID)
        self.assertEqual(
            w,
            workers.WorkerModel(
                id=self.W1_ID,
                name=self.W1_NAME,
                workerinfo=self.W1_INFO,
                paused=False,
                pause_reason=None,
                graceful=False,
                configured_on=[],
                connected_to=[],
            ),
        )

        yield self.db.workers.set_worker_graceful(self.W1_ID, True)

        w = yield self.db.workers.getWorker(self.W1_ID)
        self.assertEqual(
            w,
            workers.WorkerModel(
                id=self.W1_ID,
                name=self.W1_NAME,
                workerinfo=self.W1_INFO,
                paused=False,
                pause_reason=None,
                graceful=True,
                configured_on=[],
                connected_to=[],
            ),
        )

    @defer.inlineCallbacks
    def test_workerConfigured(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)

        # should remove builder 21, and add 22
        yield self.db.workers.deconfigureAllWorkersForMaster(masterid=10)

        yield self.db.workers.workerConfigured(workerid=30, masterid=10, builderids=[20, 22])

        w = yield self.db.workers.getWorker(30)
        self.assertEqual(
            sorted(w.configured_on, key=configuredOnKey),
            sorted(
                [
                    workers.BuilderMasterModel(builderid=20, masterid=11),
                    workers.BuilderMasterModel(builderid=20, masterid=10),
                    workers.BuilderMasterModel(builderid=22, masterid=10),
                ],
                key=configuredOnKey,
            ),
        )

    @defer.inlineCallbacks
    def test_workerConfiguredTwice(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)

        # should remove builder 21, and add 22
        yield self.db.workers.deconfigureAllWorkersForMaster(masterid=10)

        yield self.db.workers.workerConfigured(workerid=30, masterid=10, builderids=[20, 22])

        # configure again (should eat the duplicate insertion errors)
        yield self.db.workers.workerConfigured(workerid=30, masterid=10, builderids=[20, 21, 22])

        w = yield self.db.workers.getWorker(30)
        x1 = sorted(w.configured_on, key=configuredOnKey)
        x2 = sorted(
            [
                workers.BuilderMasterModel(builderid=20, masterid=11),
                workers.BuilderMasterModel(builderid=20, masterid=10),
                workers.BuilderMasterModel(builderid=21, masterid=10),
                workers.BuilderMasterModel(builderid=22, masterid=10),
            ],
            key=configuredOnKey,
        )
        self.assertEqual(x1, x2)

    @defer.inlineCallbacks
    def test_workerReConfigured(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)

        # should remove builder 21, and add 22
        yield self.db.workers.workerConfigured(workerid=30, masterid=10, builderids=[20, 22])

        w = yield self.db.workers.getWorker(30)
        w.configured_on = sorted(w.configured_on, key=configuredOnKey)
        self.assertEqual(
            w.configured_on,
            sorted(
                [
                    workers.BuilderMasterModel(builderid=20, masterid=11),
                    workers.BuilderMasterModel(builderid=20, masterid=10),
                    workers.BuilderMasterModel(builderid=22, masterid=10),
                ],
                key=configuredOnKey,
            ),
        )

    @defer.inlineCallbacks
    def test_workerReConfigured_should_not_affect_other_worker(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)

        # should remove all the builders in master 11
        yield self.db.workers.workerConfigured(workerid=30, masterid=11, builderids=[])

        w = yield self.db.workers.getWorker(30)
        x1 = sorted(w.configured_on, key=configuredOnKey)
        x2 = sorted(
            [
                workers.BuilderMasterModel(builderid=20, masterid=10),
                workers.BuilderMasterModel(builderid=21, masterid=10),
            ],
            key=configuredOnKey,
        )
        self.assertEqual(x1, x2)

        # ensure worker 31 is not affected (see GitHub issue#3392)
        w = yield self.db.workers.getWorker(31)
        x1 = sorted(w.configured_on, key=configuredOnKey)
        x2 = sorted(
            [
                workers.BuilderMasterModel(builderid=20, masterid=11),
                workers.BuilderMasterModel(builderid=22, masterid=11),
            ],
            key=configuredOnKey,
        )
        self.assertEqual(x1, x2)

    @defer.inlineCallbacks
    def test_workerUnconfigured(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)

        # should remove all builders from master 10
        yield self.db.workers.workerConfigured(workerid=30, masterid=10, builderids=[])

        w = yield self.db.workers.getWorker(30)
        w.configured_on = sorted(w.configured_on, key=configuredOnKey)
        expected = sorted(
            [workers.BuilderMasterModel(builderid=20, masterid=11)], key=configuredOnKey
        )
        self.assertEqual(w.configured_on, expected)

    @defer.inlineCallbacks
    def test_nothingConfigured(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)

        # should remove builder 21, and add 22
        yield self.db.workers.deconfigureAllWorkersForMaster(masterid=10)
        yield self.db.workers.workerConfigured(workerid=30, masterid=10, builderids=[])

        # should only keep builder for master 11
        w = yield self.db.workers.getWorker(30)
        self.assertEqual(
            sorted(w.configured_on), sorted([workers.BuilderMasterModel(builderid=20, masterid=11)])
        )

    @defer.inlineCallbacks
    def test_deconfiguredAllWorkers(self):
        yield self.db.insert_test_data(self.baseRows + self.multipleMasters)

        res = yield self.db.workers.getWorkers(masterid=11)
        self.assertEqual(len(res), 2)

        # should remove all worker configured for masterid 11
        yield self.db.workers.deconfigureAllWorkersForMaster(masterid=11)

        res = yield self.db.workers.getWorkers(masterid=11)
        self.assertEqual(len(res), 0)

    @defer.inlineCallbacks
    def test_workerConfiguredMany(self):
        manyWorkers = (
            [
                fakedb.BuilderMaster(id=1000, builderid=20, masterid=10),
            ]
            + [fakedb.Worker(id=50 + n, name='zero' + str(n)) for n in range(1000)]
            + [
                fakedb.ConfiguredWorker(id=n + 3000, workerid=50 + n, buildermasterid=1000)
                for n in range(1000)
            ]
        )
        yield self.db.insert_test_data(self.baseRows + manyWorkers)

        # should successfully remove all ConfiguredWorker rows
        with self.assertNoMaxVariables():
            yield self.db.workers.deconfigureAllWorkersForMaster(masterid=10)

        w = yield self.db.workers.getWorker(30)
        self.assertEqual(sorted(w.configured_on), [])

    @defer.inlineCallbacks
    def test_workerConfiguredManyBuilders(self):
        manyWorkers = (
            [fakedb.Builder(id=100 + n, name='a' + str(n)) for n in range(1000)]
            + [fakedb.Worker(id=50 + n, name='zero' + str(n)) for n in range(2000)]
            + [
                fakedb.BuilderMaster(id=1000 + n, builderid=100 + n, masterid=10)
                for n in range(1000)
            ]
            + [
                fakedb.ConfiguredWorker(
                    id=n + 3000, workerid=50 + n, buildermasterid=int(1000 + n / 2)
                )
                for n in range(2000)
            ]
        )
        yield self.db.insert_test_data(self.baseRows + manyWorkers)

        # should successfully remove all ConfiguredWorker rows
        with self.assertNoMaxVariables():
            yield self.db.workers.deconfigureAllWorkersForMaster(masterid=10)
        w = yield self.db.workers.getWorker(30)
        self.assertEqual(sorted(w.configured_on), [])
